"""
Test detection of duplicates in drs_tree.

"""

import os

from unittest import TestCase
import shutil
import tempfile

import gen_drs
from drslib.drs_tree import DRSTree
from drs_tree_shared import TestEg, test_dir

CHANGE_FILE = 'hfss_Amon_HadGEM2-ES_rcp45_r1i1p1_209012-209911.nc'


class TestDups(TestEg):
    def setUp(self):
        super(TestDups, self).setup()

        # Create test data
        gen_drs.write_eg1(self.tmpdir)

        # Do initial version change
        self.dt = DRSTree(self.tmpdir)
        self.dt.discover(self.incoming, activity='cmip5',
                         product='output1', institute='MOHC', model='HadCM3')
        self.pt = dt.pub_trees.values()[0]
        self.pt.do_version()

    def tearDown(self):
        shutil.rmtree(self.tmpdir)

    def _make_incoming1(self):
        # Original ingest
        gen_drs.write_listing(self.incoming, os.path.join(test_dir, 'dups1.ls'))

    def _make_incoming2(self):
        # Ingest with some new files and 2 duplicates
        gen_drs.write_listing(self.incoming, os.path.join(test_dir, 'dups2.ls'))
        
    def _make_incoming3(self):
        # As incoming2 except one of the dups differs in size
        self._make_incoming2()

        fh = open(os.path.join(self.incoming, CHANGE_FILE), 'a')
        print >>fh, 'File has grown'

    def _make_incoming4(self):
        # As incoming2 except one of the dups only differs by contents
        self._make_incoming2()

        fh = open(os.path.join(self.incoming, CHANGE_FILE), 'r+')
        fh.seek(0)
        fh.write('XXX')
        fh.close()

    #!TODO: test tracking_id detection with a real NetCDF file.



#!/usr/bin/python
## Code to identify the CMIP5 DRS "product" element based on other DRS elements and selection tables.
## Author: Martin Juckes (martin.juckes@stfc.ac.uk)
## New in this version:
##   1. cmip5_product.status no longer used
##   2. additional capability to scan previously published data
##   3. option to raise a ProductScopeexception instead of providing "False" return when arguments are inconsistent with selection tables
##   4. cmip5_product.rc has a return code on exit -- each return code coming from a unique line of code.
## 20101004 [0.8]: -- fixed bug which failed on all tables among the special cases.
##                 -- fixed scan_atomic_dataset to scan only files matching variable, table, and experiment.
##                        there is an option to turn this of (and scan all files in a directory), without which the legacy test suite will fail.
## 20101005 [0.9]: -- fixed code to deal with file time spans not aligned with calendar years.
## 20101022 [1.0]: -- added capability to extract base time from time units of netcdf files.
##                 -- fixed bug affecting 1pctCO2, abrupt4xCO2.
##                 -- fixed bug affecting response when a product change is implied.
##                 -- changed configuration file variables to be more consistent with terminology used in data request.
##                 -- made return codes more specific for OK300...
##                 -- fixed bug: offset relative to start years omitted year "1" (by starting with offset 1)
##                 -- fixed bug: piControl, decadal, aero data was using wrong offset
##                 -- changed names to be consistent with MIP tables (Co2,co2 --> CO2, volcano2010 --> volcIn2010 -- see expt_id_mappings.txt
##                 -- added code to deal with volcIn2010, aero option (previously overlooked)
## 20110201        -- debugged atomic dataset scanning to get robust estimate of nyears_submitted
##                 -- debugged support for "corresponding"
## 20110428        -- debugged logic on 1pctCO2, 3hr data -- added no_exception arcgument to select_year_list to allow ERR00##                     to be caught by calling routine.
## 20110613        -- debugged select_year_list to deal with case when data file spans time range greater than requested period
##   
version = 1.3
version_date = '20110428'
import logging
log = logging.getLogger(__name__)
import re, string

class ProductScope(Exception):
  def __init__(self,value):
     self.value = value
  def __str__(self):
     return repr(self.value)

class ddsort:
  def __init__(self,ee,k):
    self.k = k
    self.ee = ee
  def cmp(self,x,y):
    return cmp( self.ee[x][self.k], self.ee[y][self.k] )

def arg_string( var,table,expt,model,path,path_output1,path_output2):
   return '%s_%s_%s_%s__%s_%s_%s' % (var,table,expt,model,path,path_output1,path_output2)

def index_last_n_years( n, year_slices ):
    nn = len(year_slices) -1
    ny = 0
    for k in range( len(year_slices) ):
      ny += year_slices[nn][1] - year_slices[nn][0] + 1
      if ny >= n:
        return nn
      nn += -1
    return 0
      
def index_first_n_years( n, year_slices ):
    ny = 0
    for k in range( len(year_slices) ):
      ny += year_slices[k][1] - year_slices[k][0] + 1
      if ny >= n:
        return k
    return len(year_slices) -1

def index_in_list( yl, year_slices ):
    ixl = []
    ny = 0
    for k in range( len(year_slices) ):
      if year_slices[k][0] in yl or year_slices[k][1] in yl:
        ny += year_slices[k][1] - year_slices[k][0] + 1
        ixl.append(k)
    return (ixl,ny)
      
      
import shelve, os
class cmip5_product:
  def __init__(self,mip_table_shelve='sh/standard_output_mip', \
                    template='sh/template',\
                    stdo='sh/standard_output',\
                    config='ini/sample_1.ini', \
                    override_product_change_warning=False,\
                    policy_opt1='all_rel',not_ok_excpt=False):
    self.mip_sh = shelve.open( mip_table_shelve, flag='r' )
    self.tmpl = shelve.open( template, flag='r' )
    self.stdo = shelve.open( stdo, flag='r' )
    self.tmpl_keys = self.tmpl.keys()
    self.tmpl_keys.sort( ddsort(self.tmpl,0).cmp )
    self.pos_in_table = 999
    self.config = config
    self.config_exists = os.path.isfile( config )
    self.config_loaded = False
    self.uid = 'uid-unset'
    self.policy_opt1=policy_opt1
    self.last_result = ['none','none']
    self.product = 'none'
    self.ads_product = 'none'
    self.override_product_change_warning = override_product_change_warning
    self.not_ok_excpt = not_ok_excpt
    self.ScopeException = ProductScope
    self.warning = "this is a depricated variable"

  def ok(self, product, reason, rc=None):
    self.product = product
    self.reason = reason
    self.rc = rc
    log.info( '--- %s: %s: %s: %s' % ( self.rc, self.product, self.uid, self.reason ) )
    return True

  def not_ok(self, status,rc,no_except=False):
    self.reason = status
    self.rc = rc
    self.product = 'Failed'
    log.info( '--- %s: %s: %s: %s' % ( self.rc, self.product, self.uid, self.reason ) )
    if self.not_ok_excpt and not no_except:
      raise self.ScopeException( '%s:: %s' % (rc,status) )
    return False

  def scan_atomic_dataset(self,dir):
    import glob, string
    if dir[-1] != '/':
      dir += '/'
    assert os.path.isdir( dir ),'Attempt to scan a non-existent directory: %s' % dir
    if self.selective_ads_scan:
      fpat = '%s_%s_%s_%s_*.nc' % (self.var,self.table,self.model,self.expt)
    else:
      fpat = '*.nc' 
    fl = map( lambda x: string.split(x, '/')[-1], glob.glob( dir + fpat ) )
    assert len(fl) != 0,'No files matching %s found in %s' % (fpat,self.path)
    fl.sort()
    if self.path_output1 != None or self.path_output2 != None:
      if self.path_output1 != None:
        fl1b = glob.glob( self.path_output1 + '/*.nc' )
        fl1 = map( lambda x: string.split(x, '/')[-1], glob.glob( self.path_output1 + '/*.nc' ) )
      else:
        fl1 = []
      if self.path_output2 != None:
        fl2 = map( lambda x: string.split(x, '/')[-1], glob.glob( self.path_output2 + '/*.nc' ) )
      else:
        fl2 = []
      if len(fl1) == 0 and len(fl2) == 0:
        log.info( 'output1 and output2 directories are both empty: %s, %s' % (self.path_output1,self.path_output2) )
        self.new_ads = True
      fld = {}
      for f in fl:
        fld[f] = 'N'
      for f in fl1:
        assert f not in fl2, 'Files should not be in output1 and output2: %s, %s' % (self.path_output1,self.path_output2)
        if fld.has_key(f):
          fld[f] = 'R1'
        else:
          fld[f] = 'O1'
      keys = fld.keys()
      for f in fl2:
        if fld.has_key(f):
          fld[f] = 'R2'
        else:
          fld[f] = 'O2'
      fli = fl[:]
      fl = fld.keys()
      fl.sort()
      self.file_dict = fld
     
    kk = -1
    time_periods = []
    year_slices = []
    years_present = []
    time_tuples = []
    nyears = 0
    start_years = []
    for f in fl:
      bits = string.split( f[:-3], '_' )
      if len(bits) == 5:
        has_time = False
        nyears = -1
      elif len(bits) == 6:
        has_time = True
        assert bits[5][0] != '-', 'Negative date -- not supported as this version: %s' % bits[5]
        assert string.find( bits[5], '--' ) == -1, 'Negative date -- not supported as this version: %s, %s' % bits[5]
        #
        # Time parsing re-implemented by spascoe
        #
        tbits = string.split( bits[5], '-' )
        year_re = re.compile(r'(\d\d\d\d)(\d\d)?(\d\d)?(\d\d)?(\d\d)?')
        mo = year_re.match(tbits[0])
        assert mo, 'Start Date not of form yyyy[mm[dd[hh[mm]]]]: not supported: %s' % tbits[0]
        startyear = int(mo.group(1))
        assert mo.group(2) in ['01','02','03','04','05','06','07','08','09','10','11','12'], 'Unsupported date %s: yyyymm, mm should be month' % tbits[0]
        start_tuple = mo.groups()
        if len(tbits) > 1:
          mo = year_re.match(tbits[1])
          assert mo, 'End Date not of form yyyy[mm[dd[hh[mm]]]]: not supported: %s' % tbits[1]
          endyear = int(mo.group(1))
          end_tuple = mo.groups()
        else:
          endyear = startyear
          end_tuple = start_tuple
        start_years.append( startyear )
      else:
          return self.not_ok( 'filename does not match DRS: %s in %s' % (f,dir), 'ERR011' )

      if has_time:
        time_periods.append( (startyear, endyear) )
        for thisy in range( startyear, endyear+1):
          if thisy not in years_present:
             years_present.append( thisy)

        if len( year_slices ) == 0 or (startyear != year_slices[-1][0] or endyear != year_slices[-1][1]):
          year_slices.append( (startyear, endyear) )
          time_tuples.append( (start_tuple, end_tuple) )
          if len( start_tuple) ==1:
            nyears += endyear - startyear + 1
          else:
            nyears += endyear - startyear
            startm = int(start_tuple[1])
            if len( end_tuple ) == 5 and end_tuple[2] == '01' and end_tuple[3] == '00' and end_tuple[4] == '00':
              endm = int(end_tuple[1]) - 1
              if endm == 0:
                endm = 12
                nyears += -1
            else:
              endm = int(end_tuple[1])

            if endm > startm:
              nyears += 1
 
            
      elif len(fl) > 1:
          return self.not_ok( 'error: multiple files in atomic dataset with no temporal subset: %s ' % dir, 'ERR012' )
      kk +=1
      if kk == 0:
        var, mip, model, expt, ens = tuple( bits[0:5] )
        base = string.join( bits[0:5], '_' )
      elif string.join(bits[0:5],'_') != base:
          log.info( base )
          log.info( string.join(bits[0:5],'_') )
          return self.not_ok( 'error: inconsistent files in %s' % dir, 'ERR013' )
    self.nyears_submitted = len( years_present )
    self.drs = (var, mip, model, expt, ens)
    self.has_time = has_time
    self.time_periods = time_periods
    self.time_tuples = time_tuples
    self.year_slices = year_slices
    if len( time_periods ) > 0:
      self.ads_time_period = ( time_periods[0][0], time_periods[-1][1] )
    else:
      self.ads_time_period = ( -9999, -9999 )
    self.dir = dir
    self.files = fl
    self.file_start_years = start_years
    assert len(fl) == len(start_years) or len(fl) == 1, 'Apparent error in logic scanning atomic dataset'
    return True
        
  def check_var(self):
    kk = 0
    for r in self.mip_sh[self.table]:
      kk+=1
      if r[0] == self.var:
        self.vline = r[:]
        self.pos_flag = r[3]
        self.table_segment = r[4]
        if self.table_segment == None:
          self.table_segment = 0
## interim solution, to change way 1st 10 variables in day table are flagged.
        if self.pos_flag == 1:
          self.pos_in_table = 5
        else:
          self.pos_in_table = 99
        return True

    return False


  def find_rei( self,expt ):
      
    ##if expt == ['piControl','historical','amip']:
    k1 = expt
    self.category = 'centennial'
    if expt == 'piControl':
## need configuration file for branch -- if table is in [aero, day, 6hrPlev, 3hr]
      if not self.config_loaded:
        self.load_config()

## default to centennial
      if self.model not in self.cp.sections():

        log.warn( 'Model %s is not in configuration file -- category defaulting to centennial' % self.model )
        categ = 'centennial'
      else:
        opts = self.cp.options( self.model )
        if 'category' not in opts:
          log.warn( 'Category is not specified for model %s in configuration file -- defaulting to centennial' % self.model )
          categ = 'centennial'
        categ = self.cp.get( self.model, 'category' )

      if categ == 'centennial':
        k1 = 'piControl+'
      self.category = categ
      
    if k1 in self.tmpl_keys:
      self.rei = self.tmpl[k1]
      return True
    else:
      return False

  def priority(self):
      return int(float(self.vline[2]))
    
  def dimensions(self):
      return self.vline[1]

  def get_cfmip_request_spec(self):
    keys = self.stdo['cfmip'].keys()
    log.debug( 'get_cfmip_request_spec: %s' % self.rei[1] )
    if self.rei[1] not in keys:
      log.info( '%s not in keys:: %s' % (self.rei[1],str(keys)) )
      return self.ok( 'output1', 'Experiment %s not requested for cfmip' % self.rei[1], 'OK011' )
    ll = self.stdo['cfmip'][self.rei[1]]
    slice_list = ll[self.table_segment-1]
    if len(slice_list) == 0:
      self.request_spec = ('none',)
      self.nyears_requested = -1
    else:
      if slice_list[0][0] < 500:
         self.request_spec = ('listrel',)
      else:
         self.request_spec = ('list',)
#
      self.requested_years_list = []
      for s in slice_list:
         if s[1] != s[0]:
           self.requested_years_list.append( s[0] - 1 )
         for y in range( s[0], s[1]+1):
              self.requested_years_list.append( y )
         if s[1] != s[0]:
           self.requested_years_list.append( s[1] + 1 )
      self.nyears_requested = len( self.requested_years_list )
    return False
        
  def get_request_spec(self):
    tlist = self.stdo[self.request_col]
    self.requested_years_list = []
    if self.rei[0]-2 in tlist.keys():
      tli = self.rei[0]-2
      ssp = tlist[tli]
      self.request_spec = ssp
      assert ssp[0] in ['list','listrel','corres','none','all'], 'unexpected ssp[0]:: %s [%s,%s]' % (str( ssp[0] ),self.expt,self.table)
      if ssp[0] in ['list','listrel','corres']:
        nyears = 0
        for s in ssp[1:]:
          assert s[0] in ['year','slice'], 'Unexpected time specification %s' % str(s[0])
          if s[0] == 'year':
            nyears += 1
            self.requested_years_list.append( s[1] )
          elif s[0] == 'slice':
            nyears += s[2] - s[1] + 1
            for y in range( s[1], s[2]+1):
              self.requested_years_list.append( y )
        self.nyears_requested = nyears
    else:
      self.request_spec = ('none',)
      self.nyears_requested = -1

  def load_config(self):
    assert self.config_exists, 'load_config: need a valid configuration file at this point'
    if not self.config_loaded:
        import ConfigParser
        self.cp = ConfigParser.SafeConfigParser()
        self.cp.read( self.config )
        self.config_loaded = True

  def find_product(self,var,table,expt,model,path,startyear=None,verbose=False, \
                  path_output1=None, path_output2=None,selective_ads_scan=True):
    self.uid = string.join( [var,table,expt,model], '_' )
    if self.last_result[0] != arg_string( var,table,expt,model,path,path_output1,path_output2):
      self.find_product_ads(var,table,expt,model,path,verbose=verbose, path_output1=path_output1, \
                  path_output2=path_output2,selective_ads_scan=selective_ads_scan)
    self.ads_new = path_output1 == None and path_output2 == None
    
    self.ads_product = self.last_result[1]
    self.selective_ads_scan=selective_ads_scan
    assert self.last_result[1] in ['output1','output2','Failed','split','output1*','output2*','split*'], 'ERR901: Result [%s] not in accepted range' % self.last_result[1]
    if self.ads_product in ['output1','output2','output1*','output2*']:
        return True
    elif self.ads_product == 'Failed':
        return False
    else:
        if startyear in self.output1_start_years:
          self.product = 'output1'
          if self.product_change_warning:
            self.product += '*'
          return True
        elif startyear in self.file_start_years:
          self.product = 'output2'
          if self.product_change_warning:
            self.product += '*'
          return True
        else:
          return self.not_ok( 'startyear %s does not correspond to file in %s' % (startyear,path), 'ERR001' )
###################
  def find_product_ads(self,var,table,expt,model,path,verbose=False, \
                  path_output1=None, path_output2=None, selective_ads_scan=True):
    self.uid = string.join( [var,table,expt,model], '_' )
    self.selective_ads_scan=selective_ads_scan
    if self.last_result[0] == arg_string( var,table,expt,model,path,path_output1,path_output2):
      return True
    self.path_output1 = path_output1
    self.path_output2 = path_output2
    self.ads_new = path_output1 == None and path_output2 == None
## look to see if all files in the atomic dataset are treated equally'
    self.rc = 'UNSET'
    if self.find_product_step_one(var,table,expt,model,verbose=verbose):
      self.last_result = ( arg_string( var,table,expt,model,path,path_output1,path_output2), self.product )
      return True
    if self.reason == 'Experiment not identified':
      log.warn( 'Experiment [%s] not identified' % self.expt )
      return self.ok( 'output1', 'Unrequested experiment -- none replicated', 'OK001' )
    if self.reason != 'Need temporal information':
      return False
## go into deeper analysis
    self.path = path
    if self.find_product_slice():
      self.last_result = ( arg_string( var,table,expt,model,path,path_output1,path_output2), self.product )
      return True
    else:
      self.last_result = ( arg_string( var,table,expt,model,path,path_output1,path_output2), 'Failed' )
      return False
####################
####################
  def find_product_step_one(self,var,table,expt,model,verbose=False):
    if table not in self.mip_sh.keys():
      return self.not_ok( 'Bad mip table:: %s ' % table, 'ERR008' )
## offset_status has 3 levels: -1: not set, 0: set by default, 1: set using info from configuration file.
    self.offset_status = -1
    self.offset = None
    self.table = table
    self.var = var
    self.expt = expt
    self.model = model
    self.verbose = verbose
    if not self.check_var():
      return self.ok( 'output2', 'variable not requested', 'OK002' )
    if table not in ['Oyr','Omon','aero','day','6hrPlev','3hr','cfMon']:
       return self.ok( 'output1', 'Table is all in output1', 'OK013' )
    if table == 'Oyr': 
       if self.priority() == 3:
         return self.ok( 'output2', 'Table %s, priority 3' % table, 'OK003' )
       else:
         return self.ok( 'output1', 'Table %s, priority < 3' % table )
    elif table == 'Omon':
       dims = self.dimensions()
       if 'basin' in dims:
         return self.ok( 'output1', 'Table %s, with basin in dimensions' % table, 'OK004.1' )
       elif 'olevel' in dims and self.priority() >= 2:
         return self.ok( 'output2', 'Table %s, 3d, priority >=2' % table, 'OK004.2' )
       else:
         return self.ok( 'output1', 'Table %s, priority %s, dims %s' % (table,self.priority(),str(dims)), 'OK004.3' )
    elif table == 'aero' and 'alev1' not in self.dimensions():
         return self.ok( 'output1', 'Table %s, dims %s' % (table,str(self.dimensions())), 'OK005' )
    elif table == 'day' and self.pos_in_table <= 10:
         return self.ok( 'output1', 'Table %s, in first 10 variables' % table, 'OK006' )
    if verbose:
       log.info( 'need to identify experiment' )
    if not self.find_rei(expt ):
            return self.ok( 'output1','EXPERIMENT NOT IDENTIFIED', 'OK007' )
    if table in ['aero','day','6hrPlev','3hr']:
         self.request_col = {'aero':'M','day':'N', '3hr':'R', '6hrPlev':'Q'}[table]
         self.get_request_spec()
         if self.request_spec[0] in ['none','all']:
             return self.ok( 'output1', 'Request covers %s of this atomic dataset' % self.request_spec[0], 'OK008.1' )
    elif table == 'cfMon':
         if self.get_cfmip_request_spec():
             return True
         if self.request_spec[0] in ['none','all']:
             return self.ok( 'output1', 'Request covers %s of this atomic dataset' % self.request_spec[0], 'OK008.2' )
         
    return self.not_ok( 'Need temporal information', 'ERR004', no_except=True )

  def _get_file_base_year(self):
    ##try:
      ##import cdms
    ##except:
      import os, string, re
      fpath = '%s/%s' % (self.path,self.files[0])
      os.popen( 'ncdump -h %s | grep time:units > ncdump_tmp.txt' % fpath ).readlines()
      ii = open( 'ncdump_tmp.txt' ).readlines()
      if len(ii) == 0:
         raise self.ScopeException( '%s:: %s %s' % ('ERR201','ncdump of file failed:', fpath) )
      mm = re.findall( 'days since ([\d]{4})-', ii[0] )
      if len(mm) != 1:
         raise self.ScopeException( '%s:: %s' % ('ERR202','Failed to interpret ncdump output') )
      self.base_year = int(mm[0])
    
  def find_product_slice(self):
    res = 'output1'
    reason = 'default'
    if not self.scan_atomic_dataset(self.path):
       return False

    if len(self.files) == 1 and not self.has_time:
       return self.ok( 'output1', 'Singleton file with no time info', 'OK012' )

    if self.verbose:
       log.info( 'find_product_slice: years requested: %s, years submitted: %s' % (self.nyears_requested, self.nyears_submitted) )

    assert self.nyears_requested > 0, 'find_product_slice: Should not reach here with nyears_requested <= 0'
    assert self.nyears_submitted > 0, 'find_product_slice: Should not reach here with nyears_submitted <= 0'
    if self.policy_opt1 == 'all' and self.not_enough_years():
        return self.ok( 'output1',  \
            'years submitted (%s) not greater than 5 + years requested (%s)' % (self.nyears_submitted,self.nyears_requested), 'OK009.1' )

    if self.expt in ['midHolocene','lgm'] and self.table == '6hrPlev':
### take last 35 years
      return self.select_last( 35, 'output1', rc = 'OK200.01' )

    if len(self.expt) > 7 and self.expt[0:7] == 'decadal':
      time_datum = int( self.expt[7:11] )
      if time_datum in [1960, 1980, 2005]:
         requested_years = map( lambda x: x+time_datum, [10,20,30] )
      else:
         requested_years = (time_datum + 10, )
      if self.verbose:
         log.info( 'Experiment %s, time dataum %s' % (self.expt, self.time_datum ) )
      return self.select_year_list( requested_years, 'output1', rc='OK300.01' )

## override shelf spec.
    override_shelf = (self.table == '3hr' and self.expt in ['1pctCO2','abrupt4xCO2'] ) or (self.expt == 'volcIn2010' and self.table == 'aero')
       ##rspec = list( self.request_spec[0] )
       ##rspec[0] = 'listrel'
       ##self.request_spec = tuple( rspec )

## load module to deal with time slices in the variable request
    if self.request_spec[0] == 'listrel' or override_shelf:
      if self.policy_opt1 == 'all_rel' and self.not_enough_years():
         return self.ok( 'output1',  \
            'years submitted (%s) not greater than 5 + years requested (%s)' % (self.nyears_submitted,self.nyears_requested), 'OK009.2' )

      if not self.config_loaded:
        self.load_config()
      if (self.expt == 'piControl' and self.table in ['aero','day','6hrPlev'] ) or \
         (self.expt == 'esmControl' and self.table in ['aero','day'] ):
        if self.model not in self.cp.sections():
          return self.not_ok( 'Need to have model in configuration file, to specify start of %s run' % {'piControl':'historical', 'esmControl':'esmHistorical'}[self.expt], 'ERR002' )
        if self.expt == 'piControl':
          if self.category == 'centennial':
            y_pic2h = int( self.cp.get( self.model, 'branch_year_picontrol_to_historical' ) )
            y_hist0 = int( self.cp.get( self.model, 'base_year_historical' ) )
            offset = y_pic2h - y_hist0
          else:
            offset = self._get_base_year( self.expt )
            offset += -1
        else:
          y_ec2eh = int( self.cp.get( self.model, 'branch_year_esmControl_to_esmHistorical' ) )
          y_ehist0 = int( self.cp.get( self.model, 'base_year_esmHistorical' ) )
          offset = y_ec2h - y_ehist0
        requested_years = map( lambda x: x + offset, self.requested_years_list )
        return self.select_year_list( requested_years, 'output1', rc='OK300.02' )
      elif (self.table == '3hr' and self.expt in ['1pctCO2','abrupt4xCO2'] ):
        offset = self._get_base_year( self.expt )
        assert offset != None, 'offset not found for %s, %s, %s' % (self.model, self.expt, self.table)

        if self.expt == '1pctCO2':
              requested_years = map( lambda x: offset + 110 + x, range(30) )
              nrq  = 30
        else:
              requested_years = map( lambda x: offset + x, range(5) ) + \
                              map( lambda x: offset + 120 + x, range(30) )
              nrq = 35
        if self.select_year_list( requested_years, 'output1', force_complete = True, rc='OK300.03', no_except=True ):
              return True
## if not all years found, take first and last submitted as appropriate.
        if self.rc == 'ERR005':
          if self.expt == '1pctCO2':
            return self.select_last( 30, 'output1', rc='OK200.02' )
          else:
            self.select_first( 5, 'output1' )
            return self.select_last( 30, 'output1', append=True, rc='OK200.03' )
## if select_year_list has failed for some other reason (e.g. product change flagged) return false.
        else: 
          assert not self.not_ok_excpt, '%s:: Not all requested years found'
          return False

      elif (self.expt == 'piControl' and self.table == '3hr'):
        offset = self._get_base_year( '1pctCO2' )
        if offset != None:
            requested_years = map( lambda x: offset - 1 + x, self.requested_years_list )
            result = self.select_year_list( requested_years, 'output1', force_complete = True, rc='OK300.04' )
            if self.rc != 'ERR005':
                  return result
        return self.select_last( 35, 'output1', rc='OK200.04' )
          
      elif (self.expt == 'volcIn2010' and self.table == 'aero'):
        offset = self._get_base_year( self.expt, config_needed=False )
        assert offset != None, 'Failed to get base year from file for volcIn2010'
        requested_years = [offset+9,]
        return self.select_year_list( requested_years, 'output1', force_complete = True, rc='OK300.09' )
      elif self.table in ['cfMon']:
        if self.rei[1] in ['6.2a', '6.2b', '6.3-E', '6.4a', '6.4b', '6.7a', '6.7b', '6.7c']:
          return self.select_first( self.nyears_requested, 'output1' )
        elif self.rei[1] in ['5.4-l', '5.5-l', '6.1', '6.3']:
## for 6.1, 6.3 -- try to use offsets specified in configuration file
          if self.rei[1] in ['6.1','6.3']:
            assert self.expt in [ '1pctCO2', 'abrupt4xCO2'], 'ERR920: Unexpected expt (%s) at this point' % self.expt
            offset = self._get_base_year( self.expt )
          else:
            offset == None

          if offset != None:
              requested_years = map( lambda x: offset - 1 + x, self.requested_years_list )
              result =  self.select_year_list( requested_years, 'output1', force_complete = True, rc='OK300.05' )
              if self.rc != 'ERR005':
                  return result
## return years requested relative to first year if number submitted is greater than 145
          if self.nyears_submitted > 145:
             offset = self.ads_time_period[0]
             requested_years = map( lambda x: offset - 1 + x, self.requested_years_list )
             result = self.select_year_list( requested_years, 'output1', force_complete = True, rc='OK300.06' )
             if self.rc != 'ERR005':
                  return result
## return last 25 years if none of the above apply
          return self.select_last( 25, 'output1', rc='OK200.05' )
        elif self.rei[1] in ['3.1']:
          opts = self.cp.options( self.model )
          y_pic2h = int( self.cp.get( self.model, 'branch_year_picontrol_to_historical' ) )
          y_hist0 = int( self.cp.get( self.model, 'base_year_historical' ) )
          offset = y_pic2h - y_hist0
          requested_years = map( lambda x: x + offset, self.requested_years_list )
          result = self.select_year_list( requested_years, 'output1', force_complete = True, rc='OK300.07' )
          if self.rc != 'ERR005':
                  return result
          if self.table_segment == 2:
            return self.select_first( 25, 'output1' )
          else:
            self.select_first( 25, 'output1' )
            return self.select_last( 25, 'output1', append=True, rc='OK200.06' )
        else:
          return self.ok( 'output1', 'experiment not requested for cfmip', 'OK010' )

      return self.not_ok( 'listrel option not picked up for processing %s,%s' % (self.table,self.expt), 'ERR003' )
## selected time slices with absolute dates
    elif self.table in ['aero','3hr','day','6hrPlev','cfMon']:
      assert self.request_spec[0] != 'listrel', 'find_product_slice[e3]: should not have listrel here'
      return self.select_year_list( self.requested_years_list, 'output1', rc='OK300.08' )
    else:
      return self.not_ok( 'table %s not given in this method ' % self.table, 'ERR101' )
    self.reason = reason
    self.product = res
    return True

  def _check_time_slices(self):
    for k in range( len( self.year_slices ) -1 ):
       for j in range( len( self.time_tuples[k] ) ):
         if self.time_tuples[k+1][0] < self.time_tuples[k][1]:
           return self.not_ok( 'Overlapping time segments in submitted data', 'ERR009' )
## if time slices appear OK, return True
    return True

  def select_last( self, n, product, append=False, rc='OK200' ):
    if not self._check_time_slices():
      return False
    assert rc[0:5] == 'OK200', 'Bad use of select_last return code: %s' % rc
    assert product in ['output1', 'output2'], 'bad product: %s' % str(product)
    ixy = index_last_n_years( n, self.year_slices )
    if ixy == 0:
      self.product = product
      self.reason = 'all submitted years in last %s' % n
      return True
    ystart = self.year_slices[ixy][0]
    fs = []
    fns = []
    for k in range( len(self.files) ):
       if self.time_periods[k][0] >= ystart:
         fs.append( self.files[k] )
       else:
         fns.append( self.files[k] )
    if append:
      reason = self.reason + '; last %s years assigned to %s ' % (n,product)
    else:
      reason = 'last %s years assigned to %s ' % (n,product)
    return self.__assign_selected_files( fs, product, fns, reason, rc=rc, append=append )
    
  def select_first( self, n, product, rc='OK100' ):
    if not self._check_time_slices():
      return False
    assert product in ['output1', 'output2'], 'bad product: %s' % str(product)
    ixy = index_first_n_years( n, self.year_slices )
    if ixy == len(self.year_slices)-1:
      self.product = product
      self.reason = 'all submitted years in first %s' % n
      return True
    yend = self.year_slices[ixy][1]
    fs = []
    fns = []
    for k in range( len(self.files) ):
       if self.time_periods[k][1] <= yend:
         fs.append( self.files[k] )
       else:
         fns.append( self.files[k] )
    reason = 'first %s years assigned to %s ' % (n,product)
    return self.__assign_selected_files( fs, product, fns, reason, rc=rc )

  def select_year_list( self, yl, product, force_complete=False, rc='OK300', no_except=False ):
    if not self._check_time_slices():
      return False

    assert product in ['output1', 'output2'], 'bad product: %s' % str(product)
    ixy, ny = index_in_list( yl, self.year_slices )
    fs = []
    fns = []
    for k in range( len(self.files) ):
       kdone = False
       for y in yl:
         if not kdone:
           if y in range( self.time_periods[k][0],self.time_periods[k][1]+1 ):
              fs.append( self.files[k] )
              kdone = True

       if not kdone:
         if self.time_periods[k][1] in yl or self.time_periods[k][0] in yl:
           fs.append( self.files[k] )
         else:
           fns.append( self.files[k] )

    if force_complete and not ny == len(yl):
      self.res = {'fs':fs, 'fns':fns, 'ny':ny, 'yl':yl }
      return self.not_ok( 'Not all years requested have been found', 'ERR005', no_except=no_except )

    reason = 'selected years [%s/%s] assigned to %s ' % (ny,len(yl),product)
    return self.__assign_selected_files( fs, product, fns, reason, rc=rc )

  def __assign_selected_files( self, fs, product, fns, reason, rc='OK400', append=False ):
    other = {'output1':'output2', 'output2':'output1' }[product]
    if append:
      self.output1_files += { product:fs, other:fns }['output1']
      self.output2_files += { product:fs, other:fns }['output2']
    else:
      self.output1_files = { product:fs, other:fns }['output1']
      self.output2_files = { product:fs, other:fns }['output2']
    self.product_change_warning = False

    if not self.ads_new:
      nerr = [0,0,0,0]
      nt = 0
      ferr =[ [],[],[],[] ]
      ixr = { 'O1':0, 'O2':1, 'R1':2, 'R2':3 }
      for f in self.output1_files:
         if self.file_dict[f] in ['O2','R2']:
           ix = ixr[self.file_dict[f]]
           nerr[x] += 1
           ferr[x].append(f)
           nt += 1
      for f in self.output2_files:
         if self.file_dict[f] in ['O1','R1']:
           ix = ixr[self.file_dict[f]]
           nerr[ix] += 1
           ferr[ix].append(f)
           nt += 1
      self.product_change_warning = nt != 0
      if self.product_change_warning:
        log.warn( 'New data triggers change in product of already published data: O1 %s, O2 %s, R1 %s, R2 %s' % tuple(nerr) )
        emsg = ['Files in output1 moved to output2: %s', 'Files in output2 moved to output1: %s', \
                'Files in output1 updated into output2: %s', 'Files in output2 unpated into output1: %s' ]
        for i in range(4):
           if nerr[i] > 0:
             log.warn( emsg[i] % nerr[i] )
        self.output1_remove = []
        self.output2_remove = []
        self.output1_to_output2 = []
        self.output2_to_output1 = []
        fp = []
        for f in self.output1_files:
          if self.file_dict[f] in ['O2','R2']:
            if self.file_dict[f] == 'O2':
              self.output2_to_output1.append(f)
            else:
              self.output2_remove.append(f)
        for f in fp:
            self.output1_files.pop( self.output1_files.index(f) )
        fp = []
        for f in self.output2_files:
          if self.file_dict[f] in ['O1','R1']:
            if self.file_dict[f] == 'O1':
              self.output1_to_output2.append(f)
            else:
              self.output1_remove.append(f)
        for f in fp:
            self.output2_files.pop( self.output2_files.index(f) )

        if not self.override_product_change_warning:
           return self.not_ok( 'Republish data in atomic dataset or re-run with override_product_change_warning=True', 'ERR007' )

    self.output1_start_years = []
    for f in self.output1_files:
      self.output1_start_years.append( self.file_start_years[ self.files.index(f) ] )

    if self.product_change_warning:
      return self.ok( 'split*', reason, rc )
    else:
      return self.ok( 'split', reason, rc )
    
  def not_enough_years(self):
      return self.nyears_submitted < (self.nyears_requested + 6)

  def _get_base_year( self, expt, config_needed=True ):
     if self.model in self.cp.sections():
       opts = self.cp.options( self.model )
       if expt == '1pctCO2' and 'base_year_1pctCO2' in opts:
           return int( self.cp.get( self.model, 'base_year_1pctCO2' ) )
       elif expt == 'abrupt4xCO2' and 'base_year_abrupt4xCO2' in opts:
           return int( self.cp.get( self.model, 'base_year_abrupt4xCO2' ) )
       elif expt == 'piControl' and 'base_year_piControl' in opts:
           return int( self.cp.get( self.model, 'base_year_piControl' ) )

     else:
       assert not config_needed, 'Model %s not found in configuration file' % self.model
       
     if expt == self.expt:
       self._get_file_base_year()
       base = self.base_year
     else:
        base = None
     return base



# BSD Licence
# Copyright (c) 2011, Science & Technology Facilities Council (STFC)
# All rights reserved.
#
# See the LICENSE file in the source distribution of this software for
# the full license text.


import tempfile
import shutil
import sys, os
from glob import glob
from StringIO import StringIO
import datetime

from unittest import TestCase

import gen_drs
from drslib.drs_tree import DRSTree
from drslib.drs import path_to_drs, drs_to_path, DRS
from drslib import config

test_dir = os.path.dirname(__file__)

class TestEg(TestCase):
    __test__ = False

    today = int(datetime.date.today().strftime('%Y%m%d'))

    def setUp(self):
        self.tmpdir = tempfile.mkdtemp(prefix='drslib-')
        self.incoming = os.path.join(self.tmpdir, config.DEFAULT_INCOMING)

    def tearDown(self):
        shutil.rmtree(self.tmpdir)

class TestListing(TestEg):

    # Set the following in subclasses
    #   listing_file 

    def setUp(self):
        super(TestListing, self).setUp()

        listing_path = os.path.join(test_dir, self.listing_file)
        gen_drs.write_listing(self.tmpdir, listing_path)

        self.dt = DRSTree(self.tmpdir)

    def _discover(self, institute, model):
        self.dt.discover(self.incoming, activity='cmip5',
                         product='output1', 
                         institute=institute, 
                         model=model)

    def _do_version(self, pt):
        assert pt.state == pt.STATE_INITIAL
        pt.do_version()
        assert pt.state == pt.STATE_VERSIONED
        assert pt.versions.keys() == [self.today]



# BSD Licence
# Copyright (c) 2011, Science & Technology Facilities Council (STFC)
# All rights reserved.
#
# See the LICENSE file in the source distribution of this software for
# the full license text.

"""
Code to identify the CMIP5 DRS "product" element based on other DRS elements and selection tables.

Author: Martin Juckes (martin.juckes@stfc.ac.uk)

New in this version:
  1. cmip5_product.status no longer used
  2. additional capability to scan previously published data
  3. option to raise a ProductScopeexception instead of providing 
     "False" return when arguments are inconsistent with selection tables
  4. cmip5_product.rc has a return code on exit -- each return code coming 
     from a unique line of code.

20101004 [0.8]: -- fixed bug which failed on all tables among the special cases.
                -- fixed scan_atomic_dataset to scan only files matching 
                   variable, table, and experiment.  there is an option to
                   turn this of (and scan all files in a directory), 
                   without which the legacy test suite will fail.
20101005 [0.9]: -- fixed code to deal with file time spans not aligned 
                   with calendar years.

"""

import product
import init

# Exception raised by product detection failures
from product import ProductScope, cmip5_product, version, version_date



# BSD Licence
# Copyright (c) 2010, Science & Technology Facilities Council (STFC)
# All rights reserved.
#
# See the LICENSE file in the source distribution of this software for
# the full license text.

"""
A translator specific to CMIP3

The CMIP3 translation code is less complete than CMIP5.  It focuses
on translating paths to DRS instances.  

 1. CMIP3 files or paths cannot be generated from DRS instances
 2. the code doesn't use MIP tables to validate and deduce components

"""

import re

import logging
log = logging.getLogger(__name__)

import drslib.translate as T
from drslib.drs import DRS
from drslib.config import CMIP5_DRS, CMIP3_DRS

TranslationError = T.TranslationError


#!NOTE: there is no product component in CMIP3

# From Charlotte's wiki page
instmodel_map = {
    'bcc_cm1': ('CMA', 'BCC-CM1'),
    'bccr_bcm2_0': ('BCCR', 'BCM2'), 
    'cccma_cgcm3_1': ('CCCMA', 'CGCM3-1-T47'),     
    'cccma_cgcm3_1_t63': ('CCCMA','CGCM3-1-T63'),
    'cnrm_cm3': ('CNRM','CM3'),
    'miub_echo_g': ('MIUB-KMA','ECHO-G'),
    'csiro_mk3_0': ('CSIRO','MK3'),
    'csiro_mk3_5': ('CSIRO','MK3-5'),
    'gfdl_cm2_0': ('GFDL','CM2'),
    'gfdl_cm2_1': ('GFDL','CM2-1'),
    'inmcm3_0': ('INM','CM3'),
    'ipsl_cm4': ('IPSL','CM4'),
    'iap_fgoals1_0_g': ('LASG','FGOALS-G1-0'),
    'mpi_echam5': ('MPIM','ECHAM5'),
    'mri_cgcm2_3_2a': ('MRI','CGCM2-3-2'),
    'giss_aom': ('NASA','GISS-AOM'),
    'giss_model_e_h': ('NASA','GISS-EH'),
    'giss_model_e_r': ('NASA','GISS-ER'),
    'ncar_ccsm3_0': ('NCAR','CCSM3'),
    'ncar_pcm1': ('NCAR','PCM'),
    'miroc3_2_hires': ('NIES','MIROC3-2-HI'),
    'miroc3_2_medres': ('NIES','MIROC3-2-MED'),
    'ukmo_hadcm3': ('UKMO','HADCM3'),
    'ukmo_hadgem1': ('UKMO','HADGEM1'),
    'ingv_echam4': ('INGV','ECHAM4'), 
}

class InstituteModelTranslator(T.BaseComponentTranslator):
    def path_to_drs(self, context):
        instmodel = context.path_parts[CMIP3_DRS.PATH_INSTMODEL]
        
        try:
            institute, model = instmodel_map[instmodel]
        except KeyError:
            raise TranslationError('CMIP3 Institute/Model identifier %s not recognised' % instmodel)
        
        context.set_drs_component('institute', institute)
        context.set_drs_component('model', model)

    def filename_to_drs(self, context):
        pass
    
instmodel_t = InstituteModelTranslator()


class ExperimentTranslator(T.GenericComponentTranslator):
    path_i = CMIP3_DRS.PATH_EXPERIMENT
    file_i = None
    component = 'experiment'

    def _validate(self, s):
        """
        @note: No validation is done.  Any experiment is accepted
        """
        return s

experiment_t = ExperimentTranslator()

class FrequencyTranslator(T.BaseComponentTranslator):
    vocab = {'yr': 'yr', 'mo': 'mon', 'da': 'day', '3h': '3hr',
             'fixed': 'fx'}
    
    def path_to_drs(self, context):
        cmip3_freq = context.path_parts[CMIP3_DRS.PATH_FREQUENCY]
        try:
            freq = self.vocab[cmip3_freq]
        except KeyError:
            raise TranslationError('CMIP3 frequency %s not recognised' % cmip3_freq)

        context.set_drs_component('frequency', freq)

    def filename_to_drs(self, context):
        pass
    
        
frequency_t = FrequencyTranslator()


class RealmTranslator(T.GenericComponentTranslator):
    path_i = CMIP5_DRS.PATH_REALM
    file_i = None
    component = 'realm'
    # mapping cmip3 realms and variables and realms to cmip5 
    # http://metaforclimate.eu/trac/browser/cmip5q/trunk/CMIP5Outputs_to_Components
    
    # realm_map = {'cmip3Realm': ('CMIP5Realm',),}
    # atmos, ocean, land, landIce, seaIce, aerosol atmosChem, ocnBgchem
    realm_map = {'atm': ('atmos', 'aerosol', 'atmosChem','land'),
                 'ice': ('seaIce',),
                 'land': ('land','landIce'), 
                 'ocn': ('ocean',),
                 }
    # CMIP3 atmos and land variables can map onto 4 different cmip5 realms.  
    # Assume CMIP3 atm variables map onto CMIP5 atmos realm unless they appear in this dictionary 
    var_map={'atmos': {'mrsos': 'land',
                       'trsult': 'aerosol',
                       'trsul': 'aerosol',
                       'tro3': 'atmosChem',
                       },
             'land': {'sftgif': 'landIce',},
             }
                               
    def path_to_drs(self, context):
        cmip3_realm = context.path_parts[CMIP3_DRS.PATH_REALM]
        try:
            realm_list = self.realm_map[cmip3_realm]
        except KeyError:
            raise TranslationError('CMIP3 realm %s not recognised' % cmip3_realm)
        
        realm = realm_list[0]
        if len(realm_list) > 1:
            variable = context.drs.variable
            if variable is None:
                raise TranslationError('CMIP3 realm %s cannot be deduced without variable' % cmip3_realm)
            
            # Realm is replaced if it's in var_map
            try:
                realm = self.var_map[realm][variable]
            except KeyError:
                pass

        context.set_drs_component('realm', realm)

realm_t = RealmTranslator()

class EnsembleTranslator(T.BaseComponentTranslator):
    def path_to_drs(self, context):
        r_str = context.path_parts[CMIP3_DRS.PATH_ENSEMBLE]
        mo = re.match(r'run(\d+)', r_str)
        if not mo:
            raise TranslationError('Unrecognised CMIP3 ensemble identifier %s' % r_str)
        
        ensemble = (int(mo.group(1)), None, None)
        context.set_drs_component('ensemble', ensemble)

    def filename_to_drs(self, context):
        pass
            
ensemble_t = EnsembleTranslator()


class VariableTranslator(T.GenericComponentTranslator):
    path_i = CMIP3_DRS.PATH_VARIABLE
    file_i = None
    component = 'variable'

    def filename_to_drs(self, context):
        table = context.file_parts[CMIP3_DRS.FILE_TABLE]
        context.set_drs_component('table', table)

    def _validate(self, s):
        """
        @note: No validation is done.  Any variable  is accepted
        """
        return s

variable_t = VariableTranslator()


#!NOTE: No version in CMIP3

#!TODO: Subset translator

class SubsetTranslator(T.BaseComponentTranslator):
    """
    Subsets are irregular in CMIP3 so we just extract the irregular bit
    and put it in DRS.extended.
    
    """
    def filename_to_drs(self, context):
        extended = context.file_parts[CMIP3_DRS.FILE_EXTENDED]
        context.set_drs_component('extended', extended)
        
    def path_to_drs(self, context):
        pass
    
subset_t = SubsetTranslator()

class FnFixTranslator(T.BaseComponentTranslator):
    """
    Fix problem filenames in the DRS structure.

    """
    def drs_to_filepath(self, context):
        pass
        
    def path_to_drs(self, context):
        pass
    
    def filename_to_drs(self, context):
        # Detect variables containing underscores and fix.
        varbits = context.drs.variable.split('_')
        if len(varbits) > 1 and context.drs.table == varbits[-1]:
            extbits = context.drs.extended.split('_')
            #context.drs.variable = '_'.join(varbits[:-1])
            context.drs.table = extbits[0]
            context.drs.extended = '_'.join(extbits[1:])

            log.warn(('Variable containing underscore detected'
                      +'Setting variable=%s, table=%s, extended=%s')
                     % (context.drs.variable,
                        context.drs.table,
                        context.drs.extended))
                     

fnfix_t = FnFixTranslator()

class CMIP3TranslatorContext(T.TranslatorContext):
    """
    A customised context class for converting CMIP3 paths

    CMIP3 filenames are irregular so we need to override some behaviour of
    TranslatorContext.
    
    """

    #!FIXME: This regular expression doesn't allow underscores in variable
    #        names.  Unfortunately CMIP3 has them.
    _fnrexp = re.compile(r'([a-zA-Z0-9]+)_([a-zA-Z0-9]+)(?:[._-](.*))?.nc$')
    

    def __init__(self, filename=None, path=None, drs=None, table_store=None):
        assert table_store is None
        
        if path is None:
            self.path_parts = [None] * 10
        else:
            self.path_parts = path.split('/')
    
        self.file_parts = [None] * 7
        if filename is not None:
            # CMIP3 requires different filename parsing
            mo = self._fnrexp.match(filename)
            if not mo:
                raise TranslationError('Unrecognised CMIP3 filename %s' % filename)
            self.file_parts[CMIP3_DRS.PATH_VARIABLE] = mo.group(1)
            self.file_parts[CMIP3_DRS.FILE_TABLE] = mo.group(2)
            self.file_parts[CMIP3_DRS.FILE_EXTENDED] = mo.group(3)
            
        if drs is None:
            self.drs = DRS()
        else:
            self.drs = drs

class CMIP3Translator(T.Translator):
    
    ContextClass = CMIP3TranslatorContext

    translators = [instmodel_t,
                   experiment_t,
                   ensemble_t,
                   variable_t,
                   
                   # Must be processed after variable
                   realm_t,
                   frequency_t,
                   subset_t,

                   # Fix some unusual CMIP3 filenames
                   fnfix_t,
                   ]

    def init_drs(self, drs=None):
        if drs is None:
            drs = DRS()

        drs.activity = 'cmip3'
        drs.version = 1
        drs.product = 'output'

        return drs


def make_translator(prefix):
    return CMIP3Translator(prefix)




# BSD Licence
# Copyright (c) 2010, Science & Technology Facilities Council (STFC)
# All rights reserved.
#
# See the LICENSE file in the source distribution of this software for
# the full license text.

"""
A framework for translating DRS objects to and from filenames and paths.

This module is very object-heavy and over engineered.  However, it does the job.

"""

import re, os

import logging
log = logging.getLogger(__name__)

from drslib.drs import DRS
from drslib.config import CMIP5_DRS, CMIP5_CMOR_DRS

class TranslationError(Exception):
    pass


class TranslatorContext(object):
    """
    Represents a DRS URL or path being translated.

    :ivar path_parts: A list of directories following the DRS prefix
    :ivar file_parts: A list of '_'-separated parts of the filename
    :ivar drs: The DRS of interest
    :ivar table_store: The mip tables being used.

    """
    def __init__(self, filename=None, path=None, drs=None, table_store=None):
        if path is None:
            self.path_parts = [None] * 12
        else:
            self.path_parts = path.split('/')

        if filename is None:
            self.file_parts = [None] * 7
        else:
            self.file_parts = os.path.splitext(filename)[0].split('_')
            
        # Work-arround because CMOR encodes 'clim' as flag as _clim
        if self.file_parts[-1] == 'clim':
            self.file_parts[-2:] = [self.file_parts[-2] + '-clim']

        if drs is None:
            self.drs = DRS()
        else:
            self.drs = drs

        self.table_store = table_store

    def set_drs_component(self, drs_component, value):
        """
        Set a DRS component checking that the value doesn't conflict
        with any current value.

        """
        v = self.drs[drs_component]
        if v is None:
            setattr(self.drs, drs_component, value)
        else:
            if v != value:
                raise TranslationError('Conflicting value of DRS component %s' % drs_component)

    def path_to_string(self):
        self._trim_parts()

        parts = self.path_parts
        
        return os.path.join(*parts)
    
    def file_to_string(self):
        # To allow optional portions any None's are removed
        fp = [x for x in self.file_parts if x is not None]

        # Check there is at last 1 item in the list
        assert len(fp) > 0

        return '_'.join(fp)+'.nc'
    
    def to_string(self):
        """Returns the full DRS path and filename.
        """
        self._trim_parts()

        return os.path.join(self.path_to_string(), self.file_to_string())

    def _trim_parts(self):
        """Remove extranious Nones from path_parts and file_parts.
        """

        while self.path_parts[-1] is None:
            self.path_parts.pop()

        while self.file_parts[-1] is None:
            self.file_parts.pop()


class BaseComponentTranslator(object):
    """
    Each component is translated by a separate IComponentTranslator object.

    """

    def __init__(self, table_store=None):
        self.table_store = table_store

    def filename_to_drs(self, context):
        """
        Translate the relevant component of the filename to a drs component.

        :raises TranslationError: if the filename is invalid
        :returns: context

        """
        raise NotImplementedError

    def path_to_drs(self, context):
        """
        Translate the relevant component of the DRS path to a drs component.

        :raises TranslationError: if the path is invalid
        :returns: context

        """
        raise NotImplementedError


    def drs_to_filepath(self, context):
        """
        Sets context.path_parts and context.file_parts for this component.

        :returns: context

        """
        raise NotImplementedError

class GenericComponentTranslator(BaseComponentTranslator):
    """
    The simplest type of translator just validating strings

    """

    # Class attributes set in subclasses
    path_i = NotImplemented
    file_i = NotImplemented
    component = NotImplemented
    vocab = NotImplemented

    def filename_to_drs(self, context):
        if self.file_i is not None:
            try:
                s = self._validate(context.file_parts[self.file_i])
            except IndexError:
                raise TranslationError("Filename contains too few components")
            context.set_drs_component(self.component, s)


    def path_to_drs(self, context):
        if self.path_i is not None:
            s = self._validate(context.path_parts[self.path_i])
            context.set_drs_component(self.component, s)

    def drs_to_filepath(self, context):
        s = self._validate(context.drs[self.component])
        
        if self.path_i is not None:
            context.path_parts[self.path_i] = s
        if self.file_i is not None:
            context.file_parts[self.file_i] = s

    #----

    def _validate(self, s):
        if self.vocab is None:
            return s

        if s not in self.vocab:
            raise TranslationError('Component value %s not in vocabulary of component %s' % (s, self.component))

        return s





class CMORVarTranslator(BaseComponentTranslator):
    """
    CMIP Variable translator

    For CMIP5, each variable is uniquely identified by a combination
    of two strings: 1) a name associated generically with the variable
    (typically, as recorded in the netCDF file ? e.g., tas, pr, ua),
    and 2) the name of the CMOR variable table (e.g., Amon, da, aero)
    in which the variable appears. Together these two strings ensure
    that the variable is uniquely 4 defined. There are some
    applications in which the second string might be unnecessary
    (e.g., for CMIP5, this will be omitted from the directory
    structure). Note that for CMIP5 a hyphen ('-') is forbidden to be
    included anywhere in the first string.

    """

    file_var_i = CMIP5_CMOR_DRS.FILE_VARIABLE
    file_table_i = CMIP5_CMOR_DRS.FILE_TABLE
    path_var_i = CMIP5_CMOR_DRS.PATH_VARIABLE

    def filename_to_drs(self, context):
        varname = context.file_parts[self.file_var_i]
        table = context.file_parts[self.file_table_i]

        #!TODO: table checking

        context.set_drs_component('variable', varname)
        context.set_drs_component('table', table)


    def path_to_drs(self, context):
        varname = context.file_parts[self.file_var_i]
        
        #!TODO: table checking

        context.set_drs_component('variable', varname)

    def drs_to_filepath(self, context):
        context.file_parts[self.file_var_i] = context.drs.variable
        context.file_parts[self.file_table_i] = context.drs.table
        context.path_parts[self.path_var_i] = context.drs.variable


    #----

class VersionedVarTranslator(CMORVarTranslator):
    file_var_i = CMIP5_DRS.FILE_VARIABLE
    file_table_i = CMIP5_DRS.FILE_TABLE
    path_var_i = CMIP5_DRS.PATH_VARIABLE
    path_table_i = CMIP5_DRS.PATH_TABLE

    def path_to_drs(self, context):
        super(VersionedVarTranslator, self).path_to_drs(context)

        tablename = context.path_parts[self.path_table_i]

        context.set_drs_component('table', tablename)

    def drs_to_filepath(self, context):
        super(VersionedVarTranslator, self).drs_to_filepath(context)

        context.path_parts[self.path_table_i] = context.drs.table

class EnsembleTranslator(BaseComponentTranslator):
    file_i = CMIP5_CMOR_DRS.FILE_ENSEMBLE
    path_i = CMIP5_CMOR_DRS.PATH_ENSEMBLE

    def filename_to_drs(self, context):
        context.drs.ensemble = self._convert(context.file_parts[self.file_i])


    def path_to_drs(self, context):
        context.drs.ensemble = self._convert(context.path_parts[self.path_i])


    def drs_to_filepath(self, context):
        (r, i, p) = context.drs.ensemble
        a = ['r%d' % r]
        if i is not None:
            a.append('i%d' % i)
        if p is not None:
            a.append('p%d' % p)
        v = ''.join(a)
        
        context.file_parts[self.file_i] = v
        context.path_parts[self.path_i] = v


    #----

    def _convert(self, component):
        mo = re.match(r'(?:r(\d+))?(?:i(\d+))?(?:p(\d+))?', component)
        if not mo:
            raise TranslationError('Unrecognised ensemble syntax %s' % component)

        (r, i, p) = mo.groups()
        return (_int_or_none(r), _int_or_none(i), _int_or_none(p))

class VersionedEnsembleTranslator(EnsembleTranslator):
    file_i = CMIP5_DRS.FILE_ENSEMBLE
    path_i = CMIP5_DRS.PATH_ENSEMBLE


class VersionTranslator(BaseComponentTranslator):

    def filename_to_drs(self, context):
        pass

    def path_to_drs(self, context):
        context.drs.version = self._convert(context.path_parts[CMIP5_DRS.PATH_VERSION])


    def drs_to_filepath(self, context):
        if context.drs.version is not None:
            context.path_parts[CMIP5_DRS.PATH_VERSION] = 'v%d' % context.drs.version
        
    #----

    def _convert(self, component):
        mo = re.match(r'v(\d+)', component)
        if not mo:
            raise TranslationError('Unrecognised version syntax %s' % component)

        try:
            v = int(mo.group(1))
        except TypeError:
            raise TranslationError('Unrecognised version syntax %s' % component)

        return v

    

class SubsetTranslator(BaseComponentTranslator):
    """
    Currently just temporal subsets

    :cvar allow_missing_subset: A boolean value configuring whether to 
        complain if the subset is missing.

    """

    allow_missing_subset = True

    def filename_to_drs(self, context):
        try:
            v = context.file_parts[CMIP5_DRS.FILE_SUBSET]
        except IndexError:
            if self.allow_missing_subset:
                return
            else:
                raise TranslationError('Missing temporal subset')
        
        mo = re.match(r'(\d+)(?:-(\d+))?(-clim)?', v)
        if not mo:
            raise TranslationError('Unrecognised temporal subset %s' % v)

        n1, n2, clim = mo.groups()
        if clim: 
            clim=True

        context.drs.subset = (_to_date(n1), _to_date(n2), clim)


    def path_to_drs(self, component):
        pass

    def drs_to_filepath(self, context):

        if self.allow_missing_subset and context.drs.subset is None:
            return
                
        (n1, n2, clim) = context.drs.subset
        
        parts = []
        parts.append(_from_date(n1))
        if n2:
            parts.append(_from_date(n2))
        if clim:
            parts.append('clim')

        context.file_parts[CMIP5_DRS.FILE_SUBSET] = '-'.join(parts)


class Translator(object):
    """
    An abstract class for translating between filepaths and
    :class:`drslib.drs.DRS` objects.

    Concrete subclasses are returned by factory functions such as
    :mod:`drslib.cmip5.make_translator`.

    :property prefix: The prefix for all DRS paths including the
        activity.  All paths are interpreted as relative to this
        prefix.  Generated paths have this prefix added.

    :property table_store: A :class:`drslib.mip_table.MIPTableStore`
        object containing all MIP tables being used.
    
    """

    ContextClass = TranslatorContext

    def __init__(self, prefix='', table_store=None):
        self.prefix = prefix
        self.table_store = table_store

    def filename_to_drs(self, filename, context=None):
        """
        Translate a filename into a :class:`drslib.drs.DRS` object.

        Only those DRS components known from the filename will be set.

        """

        if context is None:
            context = self.ContextClass(filename=filename, drs=self.init_drs(),
                                        table_store = self.table_store)
        for t in self.translators:
            t.filename_to_drs(context)

        return context.drs

    def path_to_drs(self, path, context=None):
        """
        Translate a directory path into a :class:`drslib.drs.DRS`
        object.
        
        Only those DRS components known from the path will be set.

        """
        if context is None:
            context = self.ContextClass(path=self._split_prefix(path), drs=self.init_drs(),
                                        table_store = self.table_store)
        for t in self.translators:
            t.path_to_drs(context)

        return context.drs

    def filepath_to_drs(self, filepath, context=None):
        """
        Translate a full filepath to a :class:`drslib.drs.DRS` object.

        """
        filepath = self._split_prefix(filepath)
        path, filename = os.path.split(filepath)
        if context is None:
            context = self.ContextClass(filename=filename, path=path, drs=self.init_drs(),
                                        table_store = self.table_store)
        for t in self.translators:
            t.path_to_drs(context)
            t.filename_to_drs(context)
            
        return context.drs
            
    def drs_to_context(self, drs):
        context = self.ContextClass(drs=self.init_drs(drs), table_store = self.table_store)
        for t in self.translators:
            t.drs_to_filepath(context)

        return context

    def drs_to_filepath(self, drs):
        """
        Translate a :class:`drslib.drs.DRS` object into a full filepath.

        """
        context = self.drs_to_context(drs)

        return os.path.join(self.prefix, context.to_string())

    def drs_to_path(self, drs):
        """
        Translate a :class:`drslib.drs.DRS` object into a directory path.

        """
        context = self.drs_to_context(drs)
        
        return os.path.join(self.prefix, context.path_to_string())

    def drs_to_file(self, drs):
        """
        Translate a :class:`drslib.drs.DRS` object into a filename.

        """
        context = self.drs_to_context(drs)
        
        return context.file_to_string()

    
    def _split_prefix(self, path):
        n = len(self.prefix)
        if path[:n] == self.prefix:
            return path[n:].lstrip('/')
        else:
            log.warn('Path %s does not have prefix %s' % (path, self.prefix))
            return path

    def init_drs(self):
        """
        Returns an empty DRS instance initialised for this translator.

        This class must be overloaded in subclasses.

        """
        raise NotImplementedError


def _to_date(date_str):
    mo = re.match(r'(\d{4})(\d{2})?(\d{2})?(\d{2})?(\d{2})?', date_str)
    if not mo:
        raise ValueError()

    (y, m, d, h, mn) = mo.groups()
    return (int(y), _int_or_none(m), _int_or_none(d), _int_or_none(h), _int_or_none(mn))

def _from_date(date):
    (y, m, d, h, mn) = date

    ret = ['%04d' % y]
    if m is not None:
        ret.append('%02d' % m)
    if d is not None:
        ret.append('%02d' % d)
    if h is not None:
        ret.append('%02d' % h)
    if mn is not None:
        ret.append('%02d' % mn)
    return ''.join(ret)
    
def _int_or_none(x):
    if x is None:
        return None
    else:
        return int(x)



#!/usr/bin/env python
# BSD Licence
# Copyright (c) 2010, Science & Technology Facilities Council (STFC)
# All rights reserved.
#
# See the LICENSE file in the source distribution of this software for
# the full license text.

"""
Translate a stream of filepaths from CMIP3 to CMIP5 syntax
"""


import sys, os, re

import logging
log = logging.getLogger(__name__)

from drslib import cmip3, cmip5
from drslib.translate import TranslationError

cmip3_translator = cmip3.make_translator('')
cmip5_translator = cmip5.make_translator('')

# Redefined in main
include = None
exclude = None
dry_run = True
copy_trans = False
 
def walk_cmip3(base_path):
    """
    Walk through the CMIP3 archive
    @yield: (dirname, filenames) for terminal directories

    """
    for (dirpath, dirnames, filenames) in os.walk(base_path):

        # Store whether dirnames is empty before removing excluded
        # directories.
        is_leaf = not dirnames
        check_dirnames(dirpath, dirnames)

        if is_leaf:
            yield (dirpath, filenames)

        # This code won't be needed in Python 2.6 which has a followlinks
        # option to os.walk
        for dirname in dirnames:
            path = os.path.join(dirpath, dirname)
            if os.path.islink(path):
                log.info('Following symlink %s' % path)
                # Recursively walk the symlink
                for t in walk_cmip3(path): 
                    yield t

def check_dirnames(dirpath, dirnames):
    """
    Modify dirnames in place to remove directories according to
    include and exclude

    """
    log.debug('checking dirnames %s' % dirnames)


    for dirname in dirnames[:]:
        path = os.path.join(dirpath, dirname)
        delete = False

        for rexp in exclude:
            if rexp.match(path):
                delete = True
        for rexp in include:
            if rexp.match(path):
                delete = False

        if delete:
            log.info('Excluding directory %s' % dirname)
            dirnames.remove(dirname)

    log.debug('dirnames remaining %s' % dirnames)


def _mkdirs(name, mode=0777):
    log.info('mkdir -p %s' % name)
    if not dry_run:
        os.makedirs(name, mode)

def _copy(old, new):
    cmd = 'cp %s %s' % (old, new)
    log.info(cmd)
    if not dry_run:
        os.system(cmd)

def _rename(old, new):
    cmd = 'mv %s %s' % (old, new)
    log.info(cmd)
    if not dry_run:
        os.system(cmd)

             

def trans_files(cmip3_path, cmip5_path):
    cmip3_t = cmip3.make_translator(cmip3_path)
    cmip5_t = cmip5.make_translator(cmip5_path)

    log.info('Dry run is %s' % dry_run)
    log.info('Copying is %s' % copy_trans)

    for dirpath, filenames in walk_cmip3(cmip3_path):
        log.info('Processing directory %s' % dirpath)

        try:
            drs = cmip3_t.path_to_drs(dirpath)
            path = cmip5_t.drs_to_path(drs)
        except TranslationError, e:
            log.error('Failed to translate path %s: %s' % (dirpath, e))
            continue
        except:
            log.exception('Error translating path %s' % dirpath)
            continue

        log.info('Translating atomic dataset %s' % drs)

        if not os.path.exists(path):
            _mkdirs(path)

        for filename in filenames:

            try:
                drs2 = cmip3_t.filepath_to_drs(os.path.join(dirpath, filename))
                filename2 = cmip5_t.drs_to_file(drs2)
            except TranslationError, e:
                log.error('Failed to translate filename %s: %s' % (filename, e))
                continue

            # Sanity check
            path2 = cmip5_t.drs_to_path(drs2)
            assert path2 == path

            if copy_trans:
                _copy(os.path.join(dirpath, filename),
                      os.path.join(path, filename2))
            else:
                _rename(os.path.join(dirpath, filename),
                        os.path.join(path, filename2))

    log.info('Translation complete')


def main(argv=sys.argv):
    from optparse import OptionParser

    usage = "usage: %prog [options] cmip3_root cmip5_root"
    parser = OptionParser(usage=usage)

    parser.add_option('-i', '--include', action='append', dest='include', 
                      default=[], 
                      help='Include paths matching INCLUDE regular expression')
    parser.add_option('-e', '--exclude', action='append', dest='exclude', 
                      default=[],
                      help='Exclude paths matching EXCLUDE regular expression')

    parser.add_option('-c', '--copy', dest='copy', action='store_true', 
                     default=False,
                     help='Copy rather than move files')

    parser.add_option('-d', '--dryrun', dest='dryrun', action='store_true', 
                      default=False, 
                      help="Emit log messages but don't translate anything")

    parser.add_option('-l', '--loglevel', dest='loglevel', action='store', 
                      default='INFO',
                      help="Set logging level")

    (options, args) = parser.parse_args()
    cmip3_path, cmip5_path = args

    loglevel = getattr(logging, options.loglevel)    
    logging.basicConfig(level=loglevel,
                        format='%(asctime)s - %(levelname)s - %(message)s')

    log.info('Calling with arguments %s' % sys.argv[1:])

    if options.include:
        log.info('Include %s' % options.include)
    if options.exclude:
        log.info('Exclude %s' % options.exclude)


    # Set global variables
    global include, exclude, dry_run, copy_trans
    include = [re.compile(x) for x in options.include]
    exclude = [re.compile(x) for x in options.exclude]
    dry_run = options.dryrun
    copy_trans = options.copy

    trans_files(cmip3_path, cmip5_path)


if __name__ == '__main__':

    main(sys.argv)



# BSD Licence
# Copyright (c) 2011, Science & Technology Facilities Council (STFC)
# All rights reserved.
#
# See the LICENSE file in the source distribution of this software for
# the full license text.

"""
Bulk Generate DRS objects for testing.

"""

import random
from copy import copy
from datetime import datetime
import os, sys

from drslib.drs import DRS
from drslib import cmip5, config

from itertools import izip

# Python 2.5 compatibility
try:
    from itertools import product
except ImportError:
    def product(*args, **kwds):
        # product('ABCD', 'xy') --> Ax Ay Bx By Cx Cy Dx Dy
        # product(range(2), repeat=3) --> 000 001 010 011 100 101 110 111
        pools = map(tuple, args) * kwds.get('repeat', 1)
        result = [[]]
        for pool in pools:
            result = [x+[y] for x in result for y in pool]
        for prod in result:
            yield tuple(prod)


def random_drs(drs_template, drs_attr_values):
    """
    Create a random DRS object from a template and a dictionary
    of possible extra attribute values.

    drs_attr_values is a mapping of attribute names to either:
      1. A sequence of possible values
      2. A callable that will return a value

    """
    d = copy(drs_template)

    for attr, values in drs_attr_values:
        if hasattr(values, '__call__'):
            v = values()
        else:
            v = random.choice(values)
        setattr(d, attr, v)

    return v

def iter_random_drs(drs_template, drs_attr_values, limit=None):
    """
    Iterator that generates random DRS objects up to limit or forever.

    """
    i = 0
    while limit is None or i>=limit:
        yield random_drs(drs_template, drs_attr_values)
        i += 1

def iter_drs_template(drs_template, drs_attr_values):
    """
    Iterate over all variations of drs_template with the possible values
    from drs_attr_values.

    """
    #!NOTE: Assuming the ordering is the same
    attrs = drs_attr_values.keys()
    pos_values = drs_attr_values.values()

    for values in product(*pos_values):
        drs = copy(drs_template)
        for attr, value in izip(attrs, values):
            setattr(drs, attr, value)
        yield drs
            
def emember_range(imax, rmax=1, pmax=1):
    for i in xrange(imax):
        for r in xrange(rmax):
            for p in xrange(pmax):
                yield (i+1, r+1, p+1)

def subset_range(date1, date2, clim, n):
    #!NOTE: This uses datetime objects that won't use 360-day calendars.
    #       It therefore isn't suitable for making CMOR-correct filenames
    t_delta = (date2 - date1) / n

    for i in xrange(n):
        dt1 = date1 + t_delta*i
        dt2 = date1 + t_delta*(i+1)
        N1 = (dt1.year, dt1.month, dt1.day, dt1.hour, None)
        N2 = (dt2.year, dt2.month, dt2.day, dt2.hour, None)
        yield (N1, N2, clim)


def make_eg(**iter_template):
    template = DRS(activity='cmip5', product='output', institute='MOHC',
                   model='HadCM3', experiment='1pctto4x', 
                   frequency='day', realm='atmos', table='day',
                   )
    d = dict(ensemble=emember_range(1), subset=make_subset())
    d.update(iter_template)

    return iter_drs_template(template, d)

def make_subset(y1=2000, y2=2010, n=5):
    clim = False
    N1 = datetime(y1, 1, 1)
    N2 = datetime(y2, 1, 1)
    
    return list(subset_range(N1, N2, clim, n))

# New realm added
def eg1():
    return make_eg(variable=['tas', 'pr', 'rsus'], ensemble=emember_range(3))
def eg2():
    for r in make_eg(variable=['tas'], realm=['atmos']):
        yield r
    for r in make_eg(variable=['thetao'], table=['Omon'], realm=['ocean']):
        yield r


# New variable added to realm
def eg3_1():
    return make_eg(variable=['tas', 'pr'])
def eg3_2():
    return make_eg(variable=['rsus'])

# New files added to variable
def eg4_1():
    return make_eg(variable=['tas'], subset=make_subset()[:3])
def eg4_2():
    return make_eg(variable=['tas'], subset=make_subset()[3:])

# Files replaced in variable
def eg5_1():
    return make_eg(variable=['tas'])
def eg5_2():
    return make_eg(variable=['tas'], subset=make_subset()[:2])

def write_eg_file(filepath):
    fh = open(filepath, 'w')
    fh.write('I am %s\n' % os.path.basename(filepath))
    fh.close()


def write_eg(prefix, seq):
    """
    Create a test directory tree under prefix.
    Files are created in <prefix>/output in CMOR's
    output directory structure

    """
    trans = cmip5.make_translator(prefix, with_version=False)
    for drs in seq:
        path = trans.drs_to_filepath(drs)
        if not os.path.exists(os.path.dirname(path)):
            os.makedirs(os.path.dirname(path))
        else:
            if os.path.exists(path):
                raise RuntimeError("%s exists" % path)

        write_eg_file(path)


def write_listing(prefix, listing_file):
    """
    Create a drs-tree from a listing file
    
    """
    def it():
        for line in open(listing_file):
            line = line.strip()
            if not line or line[0] == '#':
                continue
            yield line
    write_listing_seq(prefix, it())

        
def write_listing_seq(prefix, sequence):
    """
    Create a drs-tree from a sequence.

    """
    for filename in sequence:
        if filename[0] == '/':
            raise Exception("Absolute path in listing!")

        path = os.path.normpath(filename)

        filepath = os.path.join(prefix, filename)
        if not os.path.exists(os.path.dirname(filepath)):
            os.makedirs(os.path.dirname(filepath))
        write_eg_file(filepath)


def write_eg1(prefix):
    write_eg(prefix, eg1())

def write_eg2(prefix):
    write_eg(prefix, eg2())

def write_eg3_1(prefix):
    write_eg(prefix, eg3_1())
def write_eg3_2(prefix):
    write_eg(prefix, eg3_2())

def write_eg4_1(prefix):
    write_eg(prefix, eg4_1())
def write_eg4_2(prefix):
    write_eg(prefix, eg4_2())

def write_eg5_1(prefix):
    write_eg(prefix, eg5_1())
def write_eg5_2(prefix):
    write_eg(prefix, eg5_2())



def main(argv=sys.argv):
    listing_file, outdir = sys.argv[1:]

    if os.path.exists(outdir):
        raise IOError("Directory %s already exists" % repr(outdir))
    
    write_listing(outdir, listing_file)

if __name__ == '__main__':
    main()



# BSD Licence
# Copyright (c) 2011, Science & Technology Facilities Council (STFC)
# All rights reserved.
#
# See the LICENSE file in the source distribution of this software for
# the full license text.

"""
Command-line access to drslib

"""

import sys, os

from optparse import OptionParser
from ConfigParser import NoSectionError, NoOptionError

from drslib.drs_tree import DRSTree
from drslib import config
from drslib.drs import DRS

from drslib import p_cmip5

import logging
log = logging.getLogger(__name__)

usage = """\
usage: %prog [command] [options] [drs-pattern]
       %prog --help

command:
  list            list publication-level datasets
  todo            show file operations pending for the next version
  upgrade         make changes to the selected datasets to upgrade to the next version.
  mapfile         make a mapfile of the selected dataset
  history         list all versions of the selected dataset
  init            initialise CMIP5 data for product detection
  diff            list differences between versions or between a version and the todo list 

drs-pattern:
  A dataset identifier in '.'-separated notation using '%' for wildcards

Run %prog --help for full list of options.
"""



def make_parser():
    op = OptionParser(usage=usage)

    op.add_option('-R', '--root', action='store',
                  help='Root directory of the DRS tree')
    op.add_option('-I', '--incoming', action='store',
                  help='Incoming directory for DRS files.  Defaults to <root>/%s' % config.DEFAULT_INCOMING)
    for attr in ['activity', 'product', 'institute', 'model', 'experiment', 
                 'frequency', 'realm']:
        op.add_option('-%s'%attr[0], '--%s'% attr, action='store',
                      help='Set DRS attribute %s for dataset discovery'%attr)

    op.add_option('-v', '--version', action='store',
                  help='Force version upgrades to this version')

    op.add_option('-P', '--profile', action='store',
                  metavar='FILE',
                  help='Profile the script exectuion into FILE')

    # p_cmip5 options
    op.add_option('--detect-product', action='store_true', 
                  help='Automatically detect the DRS product of incoming data')
    op.add_option('--shelve-dir', action='store',
                  help='Location of the p_cmip5 data directory')
    op.add_option('--p-cmip5-config', action='store',
                  help='Location of model-specific configuration file for p_cmip5')

    op.add_option('-M', '--move-cmd', action='store',
                  help='Set the command used to move files into the DRS structure')

    return op

class Command(object):
    def __init__(self, opts, args):
        self.opts = opts
        self.args = args
        self.shelve_dir = None
        self.p_cmip5_config = None
        self.drs_root = None
        self.drs_tree = None

        self.make_drs_tree()

    def _config_p_cmip5(self):
        """
        Ensure self.shelve_dir is set.  This is required for InitCommand
        and any command that uses p_cmip5.

        """
        self.shelve_dir = self.opts.shelve_dir
        if self.shelve_dir is None:
            try:
                self.shelve_dir = config.config.get('p_cmip5', 'shelve-dir')
            except NoSectionError:
                raise Exception("Shelve directory not specified.  Please use --shelve-dir or set shelve_dir via metaconfig")

    def _setup_p_cmip5(self):
        """
        Instantiate the p_cmip5.cmip5_product object ready for deducing
        the product component.

        """
        
        shelves = p_cmip5.init._find_shelves(self.shelve_dir)
    
        self.p_cmip5_config = self.opts.p_cmip5_config
        if self.p_cmip5_config is None:
            try:
                self.p_cmip5_config = config.config.get('p_cmip5', 'config')
            except (NoSectionError, NoOptionError):
                raise Exception("p_cmip5 configuration file not specified.  Please use --p-cmip5-config or set via metaconfig")

        self.drs_tree.set_p_cmip5(p_cmip5.product.cmip5_product(
                mip_table_shelve=shelves['stdo_mip'],
                template=shelves['template'],
                stdo=shelves['stdo'],
                config=self.p_cmip5_config,
                not_ok_excpt=True))


    def make_drs_tree(self):
        if self.opts.root:
            self.drs_root = self.opts.root
        else:
            try:
                self.drs_root = config.drs_defaults['root']
            except KeyError:
                raise Exception('drs-root not defined')

        if self.opts.incoming:
            incoming = self.opts.incoming
        else:
            try:
                incoming = config.drs_defaults['incoming']
            except KeyError:
                incoming = os.path.join(self.drs_root, config.DEFAULT_INCOMING)

        self.drs_tree = DRSTree(self.drs_root)

        if self.opts.move_cmd:
            self.drs_tree.set_move_cmd(self.opts.move_cmd)


        kwargs = {}
        for attr in ['activity', 'product', 'institute', 'model', 'experiment', 
                     'frequency', 'realm', 'ensemble']:
            try:
                val = getattr(self.opts, attr)
                # val may be there but None
                if val is None:
                    raise AttributeError
            except AttributeError:
                val = config.drs_defaults.get(attr)

            kwargs[attr] = val

        # Get the template DRS from args
        if self.args:
            dataset_id = self.args.pop(0)
            drs = DRS.from_dataset_id(dataset_id, **kwargs)
        else:
            drs = DRS(**kwargs)

        # Product detection
        if self.opts.detect_product:
            self._config_p_cmip5()
            self._setup_p_cmip5()

        self.drs_tree.discover(incoming, **drs)

    def do(self):
        raise NotImplementedError("Unimplemented command")
    

    def print_header(self):
        print """\
==============================================================================
DRS Tree at %s
------------------------------------------------------------------------------\
""" % self.drs_root

    def print_sep(self):
        print """\
------------------------------------------------------------------------------\
"""

    def print_footer(self):
        print """\
==============================================================================\
"""



class ListCommand(Command):
    def do(self):
        self.print_header()

        to_upgrade = 0
        for k in sorted(self.drs_tree.pub_trees):
            pt = self.drs_tree.pub_trees[k]
            todo = pt.count_todo()
            if todo:
                state_msg = '%d:%d %d:%d' % (pt.count(), pt.size(), todo, pt.todo_size())
            else:
                state_msg = '%d:%d' % (pt.count(), pt.size())
            if pt.state != pt.STATE_VERSIONED:
                to_upgrade += 1
            #!TODO: print update summary
            print '%-70s  %s' % (pt.version_drs().to_dataset_id(with_version=True), state_msg)
    
        if self.drs_tree.incomplete:
            self.print_sep()
            print 'Incompletely specified incoming datasets'
            self.print_sep()
            for dataset_id in sorted(self.drs_tree.incomplete_dataset_ids()):
                print '%-70s' % dataset_id

        if to_upgrade:
            self.print_sep()
            print '%d datasets awaiting upgrade' % to_upgrade
        self.print_footer()

class TodoCommand(Command):
    def do(self):
        self.print_header()
        first = True
        for k in sorted(self.drs_tree.pub_trees):
            pt = self.drs_tree.pub_trees[k]

            if pt.count_todo() == 0:
                if not first: 
                    self.print_sep()
                print 'Nothing todo for %s' % pt.drs.to_dataset_id()
                first = False
                continue

            if self.opts.version:
                next_version = int(self.opts.version)
            else:
                next_version = pt._next_version()

            todos = pt.list_todo(next_version)
            if not first:
                self.print_sep()
            print "Publisher Tree %s todo for version %d" % (pt.drs.to_dataset_id(),
                                                             next_version)
            first = False
            print
            print '\n'.join(todos)
        self.print_footer()

class UpgradeCommand(Command):
    def do(self):
        self.print_header()

        for k in sorted(self.drs_tree.pub_trees):
            pt = self.drs_tree.pub_trees[k]
            if self.opts.version:
                next_version = int(self.opts.version)
            else:
                next_version = pt._next_version()

            if pt.state == pt.STATE_VERSIONED:
                print 'Publisher Tree %s has no pending upgrades' % pt.drs.to_dataset_id()
            else:
                print ('Upgrading %s to version %d ...' % (pt.drs.to_dataset_id(), next_version)),
                to_process = pt.count_todo()
                pt.do_version(next_version)
                print 'done %d' % to_process

        self.print_footer()

class MapfileCommand(Command):
    def do(self):
        """
        Generate a mapfile from the selection.  The selection must be for
        only 1 realm-tree.

        """

        if len(self.drs_tree.pub_trees) != 1:
            raise Exception("You must select 1 dataset to create a mapfile.  %d selected" %
                            len(self.drs_tree.pub_trees))

        if len(self.drs_tree.pub_trees) == 0:
            raise Exception("No datasets selected")

        pt = self.drs_tree.pub_trees.values()[0]

        #!TODO: better argument handling
        if self.args:
            version = int(self.args[0])
        else:
            version = pt.latest

        if version not in pt.versions:
            log.warning("PublisherTree %s has no version %d, skipping" % (pt.drs.to_dataset_id(), version))
        else:
            #!TODO: Alternative to stdout?
            pt.version_to_mapfile(version)

class HistoryCommand(Command):
    def do(self):
        """
        List all versions of a selected dataset.

        """
        if len(self.drs_tree.pub_trees) != 1:
            raise Exception("You must select 1 dataset to list history.  %d selected" % len(self.drs_tree.pub_trees))
        pt = self.drs_tree.pub_trees.values()[0]
        
        self.print_header()
        print "History of %s" % pt.drs.to_dataset_id()
        self.print_sep()
        for version in sorted(pt.versions, reverse=True):
            vdrs = DRS(pt.drs, version=version)
            print vdrs.to_dataset_id(with_version=True)
        self.print_footer()
            

class InitCommand(Command):
    def make_drs_tree(self):
        """No need to initialise the drs tree for this command.
        """
        pass

    def do(self):
        self._config_p_cmip5()

        from drslib.p_cmip5.init import init
        init(self.shelve_dir, config.table_path)

        print "CMIP5 configuration data written to %s" % repr(self.shelve_dir)



class DiffCommand(Command):
    """
    Accepts 0-2 arguments.  If no arguments given lists the diff between
    the latest version and the todo list.  If 1 given lists  the diff
    between that version and the todo list.  If 2 given lists the diff
    between these versions.

    """
    def do(self):
        if len(self.drs_tree.pub_trees) != 1:
            raise Exception("You must select 1 dataset to view differences. %d selected" %
                            len(self.drs_tree.pub_trees))

        if len(self.drs_tree.pub_trees) == 0:
            raise Exception("No datasets selected")

        pt =self.drs_tree.pub_trees.values()[0]

        #!TODO: better argument handling
        args = self.args[:]
        if not args:
            v1 = pt.latest
            v2 = None
        else:
            v1 = int(args.pop(0))
            if args:
                v2 = int(args.pop(0))
            else:
                v2 = None
        
        # Yields DIFF_STATE, file1, file2

        self.print_header()
        if v2:
            v2_msg = v2
        else:
            v2_msg = 'todo list'
        print 'Diff between %s and %s' % (v1, v2_msg)
        self.print_sep()

        #!FIXME: Just compare file sizes at the moment!
        for diff_type, f1, f2 in pt.diff_version(v1, v2):
            filename = os.path.basename(f1 or f2)
            if diff_type == pt.DIFF_NONE:
                continue
            # Don't compare by path if using the todo list
            elif (v2 is not None) and (diff_type & pt.DIFF_PATH == pt.DIFF_PATH):
                print 'PATH\t\t%s' % filename
            elif diff_type & pt.DIFF_SIZE == pt.DIFF_SIZE:
                print 'SIZE\t\t%s' % filename
            elif diff_type & pt.DIFF_TRACKING_ID == pt.DIFF_TRACKING_ID:
                print 'TRACKING_ID\t%s' % filename
            elif diff_type & pt.DIFF_V1_ONLY == pt.DIFF_V1_ONLY:
                print '%s\t\t%s' % (v1, filename)
            elif diff_type & pt.DIFF_V2_ONLY == pt.DIFF_V2_ONLY:
                print '%s\t\t%s' % (v2_msg, filename)
            else:
                assert False
                
        self.print_footer()

def run(op, command, opts, args):
    if command == 'list':
        c = ListCommand(opts, args)
    elif command == 'todo':
        c = TodoCommand(opts, args)
    elif command == 'upgrade':
        c = UpgradeCommand(opts, args)
    elif command == 'mapfile':
        c = MapfileCommand(opts, args)
    elif command == 'history':
        c = HistoryCommand(opts, args)
    elif command == 'init':
        c = InitCommand(opts, args)
    elif command == 'diff':
        c = DiffCommand(opts, args)
    else:
        op.error("Unrecognised command %s" % command)

    c.do()


def main(argv=sys.argv):

    op = make_parser()

    try:
        command = argv[1]
    except IndexError:
        op.error("command not specified")

    #!FIXME: better global vs. per-command help
    if command in ['-h', '--help']:
        opts, args = op.parse_args(argv[1:2])
    else:
        opts, args = op.parse_args(argv[2:])
    
    if opts.profile:
        import cProfile
        cProfile.runctx('run(op, command, opts, args)', globals(), locals(), opts.profile)
    else:
        return run(op, command, opts, args)

if __name__ == '__main__':
    main()



# BSD Licence
# Copyright (c) 2010, Science & Technology Facilities Council (STFC)
# All rights reserved.
#
# See the LICENSE file in the source distribution of this software for
# the full license text.

"""

Translate CMIP5 DRS filenames and paths to and from DRS objects with
consistency checking against CMIP5 MIP tables and between the filename
and path portions of filepaths.

"""

import drslib.translate as T
from drslib import config
from drslib.mip_table import read_model_table

import logging
log = logging.getLogger(__name__)

class ProductTranslator(T.GenericComponentTranslator):
    path_i = T.CMIP5_DRS.PATH_PRODUCT
    file_i = None
    component = 'product'
    vocab = None


#!TODO: Get official list.  This is based on Karl's spreadsheet and some educated guesses

# CCSR, CNRM, CSIRO, GFDL, INM, IPSL, LASG, MOHC, MPI-M, MRI, NCAR, NCC, NIMR


#!TODO: make this configurable with metaconfig
model_institute_map = read_model_table(config.model_table)        
cmip3_models = {
    'BCC-CM1': 'CMA',
    'BCM2': 'BCCR',
    'CGCM3-1-T47': 'CCCMA',
    'CGCM3-1-T63': 'CCCMA',
    'CM3': 'CNRM',
    'ECHO-G': 'MIUB-KMA',
    'MK3': 'CSIRO',
    'MK3-5': 'CSIRO',
    'CM2': 'GFDL',
    'CM2-1': 'GFDL',
    'CM3': 'INM',
    'CM4': 'IPSL',
    'FGOALS-G1-0': 'LASG',
    'ECHAM5': 'MPIM',
    'CGCM2-3-2': 'MRI',
    'GISS-AOM': 'NASA',
    'GISS-EH': 'NASA',
    'GISS-ER': 'NASA',
    'CCSM3': 'NCAR',
    'PCM': 'NCAR',
    'MIROC3-2-HI': 'NIES',
    'MIROC3-2-MED': 'NIES',
    'HADCM3': 'UKMO',
    'HADGEM1': 'UKMO',
    'ECHAM4': 'INGV',
    
    # Models from CMOR test suite
    'GICCM1': 'TEST',

    # Models in test listings contributed from MPI
    'ECHAM6-MPIOM-HR': 'MPI-M',
    'ECHAM6-MPIOM-LR': 'MPI-M',
}
for k in cmip3_models:
    if k in model_institute_map:
        raise Exception("Duplicate model %s" % k)
    model_institute_map[k] = cmip3_models[k]

class InstituteTranslator(T.GenericComponentTranslator):
    path_i = T.CMIP5_DRS.PATH_INSTITUTE
    file_i = None
    component = 'institute'
    vocab = model_institute_map.values()

    def filename_to_drs(self, context):
        context.drs.institute = self._deduce_institute(context)

    def drs_to_filepath(self, context):
        if context.drs.institute is None:
            context.drs.institute = self._deduce_institute(context)

        super(InstituteTranslator, self).drs_to_filepath(context)        

    #----

    def _deduce_institute(self, context):
        model = context.drs.model
        if context.drs.institute:
            return context.drs.institute
        try:
            return model_institute_map[model]
        except KeyError:
            log.warn('Institute translation requires model to be known')
            return None


    # Allow all institutes
    def _validate(self, s):
        return s


#!TODO: Not official identifiers
class ModelTranslator(T.GenericComponentTranslator):
    path_i = T.CMIP5_DRS.PATH_MODEL
    file_i = T.CMIP5_DRS.FILE_MODEL
    component = 'model'
    vocab = model_institute_map.keys()

    def _validate(self, s):
        # Demote validation errors to a warning.
        try:
            return super(ModelTranslator, self)._validate(s)
        except T.TranslationError, e:
            log.warning('Model validation error: %s', e)
        return s

model_t = ModelTranslator()

class ExperimentTranslator(T.GenericComponentTranslator):
    path_i = T.CMIP5_DRS.PATH_EXPERIMENT
    file_i = T.CMIP5_DRS.FILE_EXPERIMENT
    component = 'experiment'
    #!NOTE: Set CMIP3 and decadal experiments in metaconfig.
    vocab = set()

    def __init__(self, table_store):
        super(ExperimentTranslator, self).__init__(table_store)

        # Get valid experiment ids from MIP tables
        for table in self.table_store.tables.values():
            self.vocab.update(table.experiments)

        # Get valid experiment ids from metaconfig
        self.vocab.update(config.experiments)

class FrequencyTranslator(T.GenericComponentTranslator):
    path_i = T.CMIP5_DRS.PATH_FREQUENCY
    file_i = None
    component = 'frequency'

    def __init__(self, table_store):
        super(FrequencyTranslator, self).__init__(table_store)

        self.vocab = set()
        for table in self.table_store.tables.values():
            try:
                self.vocab.add(table.frequency)
            except AttributeError:
                pass

    def filename_to_drs(self, context):
        context.drs.frequency = self._deduce_freq(context)

    def drs_to_filepath(self, context):
        # If context.drs.frequency is None it could be deduced from the MIP table
        if context.drs.frequency is None:
            context.drs.frequncy = self._deduce_freq(context)

        return super(FrequencyTranslator, self).drs_to_filepath(context)
            
    #----

    def _deduce_freq(self, context):
        # Read frequency from MIP table
        table = context.drs.table
        variable = context.drs.variable
        if (table is None) or (variable is None):
            raise T.TranslationError('Frequency translation requires table and variable to be known')

        return context.table_store.get_global_attr(table, 'frequency')
        

#!TODO: Get this information from CMIP tables
class RealmTranslator(T.GenericComponentTranslator):
    path_i = T.CMIP5_DRS.PATH_REALM
    file_i = None
    component = 'realm'

    def __init__(self, table_store):
        super(RealmTranslator, self).__init__(table_store)

        # Extract valid realms from the MIP tables
        self.vocab = set()
        for table in table_store.tables.values():
            for var in table.variables:
                try:
                    realms = table.get_variable_attr(var, 'modeling_realm')[0]
                except AttributeError:
                    pass
                else:
                    realms = realms.split()
                    self.vocab.update(realms)

    def _validate(self, s):
        # Multi-valued realms.  self._validate automatically selects
        # the first realm to put in the DRS syntax.
        #!TODO: smarter algorithm for deciding main realm
        if ' ' in s:
            s = s.split(' ')[0]

        return super(RealmTranslator, self)._validate(s)

    def filename_to_drs(self, context):
        try:
            context.drs.realm = self._deduce_realm(context)
        except T.TranslationError:
            log.warning("Realm translation not possible.  You must provide the realm manually")

    def drs_to_filepath(self, context):
        # If context.drs.realm is None it could be deduced from the MIP table
        if context.drs.realm is None:
            context.drs.realm = self._deduce_realm(context)

        return super(RealmTranslator, self).drs_to_filepath(context)


    #----

    def _deduce_realm(self, context):
        # Read realm from MIP table
        table = context.drs.table
        variable = context.drs.variable
        if (table is None) or (variable is None):
            raise T.TranslationError('Realm translation requires table and variable to be known')

        try:
            val = context.table_store.get_variable_attr(table, variable, 'modeling_realm')
        except ValueError:
            raise T.TranslationError

        return self._validate(val)



class ExtendedTranslator(T.BaseComponentTranslator):
    """
    The extended DRS component is only used when converting DRS->filename.
    It is needed for CMIP3 conversions.
    
    """
    def drs_to_filepath(self, context):
        context.file_parts[T.CMIP5_DRS.FILE_EXTENDED] = context.drs.extended
        
    def path_to_drs(self, context):
        pass
    
    def filename_to_drs(self, context):
        pass
    



class CMIP5Translator(T.Translator):
    def init_drs(self, drs=None):
        if drs is None:
            drs = T.DRS()

        if drs.activity is None:
            drs.activity = 'cmip5'

        return drs

_table_store = None
def get_table_store():
    """
    Return a :class:`drslib.mip_table.MIPTableStore` object
    containing the CMIP5 MIP tables available.

    """
    global _table_store
    from drslib.mip_table import MIPTableStore

    if _table_store is None:
        _table_store = MIPTableStore('%s/%s*' % (config.table_path, 
                                                 config.table_prefix))

    return _table_store

def make_translator(prefix, with_version=True, table_store=None):
    """
    Return a :class:`drslib.translate.Translator` object for
    translating filepaths to and from ``DRS`` instances.

    :param prefix: The path to the root of the DRS tree.  This should
        point to the DRS ``activity`` directory.

    :param with_version: If ``True`` the translator will include a
        version directory in filesystem paths, otherwise it reflects
        the output structure of CMOR.

    :param table_store:: Override default table store.

    """

    if table_store is None:
        table_store = get_table_store()

    t = CMIP5Translator(prefix, table_store)
    t.translators = [
        ProductTranslator(table_store),
        ModelTranslator(table_store),

        # Must follow model_t
        InstituteTranslator(table_store),
        ExperimentTranslator(table_store),
        ]

    if with_version:
        t.translators += [
            T.VersionedEnsembleTranslator(table_store),
            T.VersionedVarTranslator(table_store),
            ]
    else:
        t.translators += [
            T.EnsembleTranslator(table_store),
            T.CMORVarTranslator(table_store),
            ]
        
    t.translators += [
        # Must be processed after variable
        RealmTranslator(table_store),
        FrequencyTranslator(table_store),
        ]

    if with_version:
        t.translators.append(T.VersionTranslator(table_store))

    t.translators += [
        T.SubsetTranslator(table_store),
        ExtendedTranslator(table_store),
        ]


    return t






# BSD Licence
# Copyright (c) 2010, Science & Technology Facilities Council (STFC)
# All rights reserved.
#
# See the LICENSE file in the source distribution of this software for
# the full license text.

'''

The drs module contains a minimal model class for DRS information
and some utility functions for converting filesystem paths to and from
DRS objects.

More sophisticated conversions can be done with the
:mod:`drslib.translate` and :mod:`drslib.cmip5` modules.

'''

import os
import itertools
import re

import logging
log = logging.getLogger(__name__)

DRS_ATTRS = ['activity', 'product', 'institute', 'model', 'experiment', 'frequency', 
             'realm', 'table', 'ensemble', 'version', 'variable', 'subset', 'extended']
PUB_ATTRS = ['activity', 'product', 'institute', 'model', 'experiment', 'frequency', 
             'realm', 'table', 'ensemble', ]


class DRS(dict):
    """
    Represents a DRS entry.  DRS objects are dictionaries where DRS
    components are also exposed as attributes.  Therefore you can get/set
    DRS components using dictionary or attribute notation.

    In combination with the translator machinary, this class maintains
    consistency between the path and filename portion of the DRS.

    :ivar activity: string
    :ivar product: string
    :ivar institute: string
    :ivar model: string
    :ivar experiment: string
    :ivar frequency: string
    :ivar realm: string
    :ivar variable: string
    :ivar table: string of None
    :ivar ensemble: (r, i, p)
    :ivar version: integer
    :ivar subset: (N1, N2, clim) where N1 and N2 are (y, m, d, h, mn) 
        and clim is boolean
    :ivar extended: A string containing miscellaneous stuff.  Useful for
        representing irregular CMIP3 files

    """

    

    def __init__(self, *argv, **kwargs):
        """
        Instantiate a DRS object with a set of DRS component values.

        >>> mydrs = DRS(activity='cmip5', product='output', model='HadGEM1',
        ...             experiment='1pctto4x', variable='tas')
        <DRS activity="cmip5" product="output" model="HadGEM1" ...>

        :param argv: If not () should be a DRS object to instantiate from
        :param kwargs: DRS component values.

        """

        # Initialise all components as None
        for attr in DRS_ATTRS:
            self[attr] = None

        # Check only DRS components are used
        for kw in kwargs:
            if kw not in DRS_ATTRS:
                raise KeyError("Keyword %s is not a DRS component" % repr(kw))

        # Use dict flexible instantiation
        super(DRS, self).__init__(*argv, **kwargs)


    def __getattr__(self, attr):
        if attr in DRS_ATTRS:
            return self[attr]
        else:
            raise AttributeError('%s object has no attribute %s' % 
                                 (repr(type(self).__name__), repr(attr)))

    def __setattr__(self, attr, value):
        if attr in DRS_ATTRS:
            self[attr] = value
        else:
            raise AttributeError('%s is not a DRS component' % repr(attr))

    def is_complete(self):
        """Returns boolean to indicate if all components are specified.
        
        Returns ``True`` if all components except ``extended`` have a value.

        """

        for attr in DRS_ATTRS:
            if attr is 'extended':
                continue
            if self.get(attr, None) is None:
                return False

        return True

    def is_publish_level(self):
        """Returns boolian to indicate if the all publish-level components are
        specified.

        """
        for attr in PUB_ATTRS:
            if self.get(attr, None) is None:
                return False

        return True

    def __repr__(self):
        kws = []
        for attr in DRS_ATTRS:
            kws.append(self._encode_component(attr))

        # Remove trailing '%' from components
        while kws[-1] == '%':
            kws.pop(-1)

        return '<DRS %s>' % '.'.join(kws)

    def _encode_component(self, attr):
        """
        Encode a DRS component as a string.  Components that are None
        are encoded as '%'.

        """
        from drslib.translate import _to_date, _from_date

        #!TODO: this code overlaps serialisation code in translate.py
        if self[attr] is None:
            val = '%'
        elif attr is 'ensemble':
            val = self._encode_ensemble()
        elif attr is 'version':
            val = 'v%d' % self.version
        elif attr is 'subset':
            N1, N2, clim = self.subset
            if clim:
                val = '%s-%s-clim' % (_from_date(N1), _from_date(N2))
            else:
                val = '%s-%s' % (_from_date(N1), _from_date(N2))
        else:
            val = self[attr]

        return val

    def _encode_ensemble(self):
        r, i, p = self.ensemble
        ret = 'r%d' % r
        if i is not None:
            ret += 'i%d' % i
            if p is not None:
                ret += 'p%d' % p

        return ret

    def to_dataset_id(self, with_version=False):
        """
        Return the esgpublish dataset_id for this drs object.
        
        If version is not None and with_version=True the version is included.

        """
        parts = [self._encode_component(x) for x in PUB_ATTRS]
        if self.version and with_version:
            parts.append(self._encode_component('version'))
        return '.'.join(parts)

    @classmethod
    def from_dataset_id(klass, dataset_id, **components):
        """
        Return a DRS object fro a ESG Publisher dataset_id.

        If the dataset_id contains less than 10 components all trailing
        components are set to None.  Any component of value '%' is set to None

        E.g.
        >>> drs = DRS.from_dataset_id('cmip5.output.MOHC.%.rpc45')
        >>> drs.institute, drs.model, drs.experiment, drs.realm
        ('MOHC', None, 'rpc45', None)

        """

        parts = dataset_id.split('.')
        for attr, val in itertools.izip(DRS_ATTRS, parts):
            if val is '%':
                continue
            if attr is 'ensemble':
                r, i, p = re.match(r'r(\d+)i(\d+)p(\d+)', val).groups()
                components[attr] = (int(r), int(i), int(p))
            elif attr is 'version':
                v = re.match(r'v(\d+)', val).group(1)
                components[attr] = int(v)
                # Don't process after version
                break
            else:
                components[attr] = val
                   
        return klass(**components)
            


#--------------------------------------------------------------------------
# A more lightweight way of getting the DRS attributes from a path.
# This is effective for the path part of a DRS path but doesn't verify
# or parse the filename
#

def path_to_drs(drs_root, path, activity=None):
    """
    Create a :class:`DRS` object from a filesystem path.

    This function is more lightweight than using :mod:`drslib.translator`
    but only works for the parts of the DRS explicitly represented in
    a path.

    :param drs_root: The root of the DRS tree.  
        This should point to the *activity* directory

    :param path: The path to convert.  This is either an absolute path
        or is relative to the current working directory.

    """

    nroot = drs_root.rstrip('/') + '/'
    relpath = os.path.normpath(path[len(nroot):])

    p = relpath.split('/')
    attrs = ['product', 'institute', 'model', 'experiment',
             'frequency', 'realm', 'table', 'ensemble'] 
    drs = DRS(activity=activity)
    for val, attr in itertools.izip(p, attrs):
        if attr == 'ensemble':
            mo = re.match(r'r(\d+)i(\d+)p(\d+)', val)
            drs[attr] = tuple(int(x) for x in mo.groups())
        else:
            drs[attr] = val

    log.debug('%s => %s' % (repr(path), drs))

    return drs
    
    
def drs_to_path(drs_root, drs):
    """
    Returns a directory path from a :class:`DRS` object.  Any DRS component
    that is set to None will result in a wildcard '*' element in the path.

    This function does not take into account of MIP tables of filenames.
    
    :param drs_root: The root of the DRS tree.  This should point to
        the *activity* directory
    
    :param drs: The :class:`DRS` object from which to generate the path

    """
    attrs = ['product', 'institute', 'model', 'experiment',
             'frequency', 'realm', 'table', 'ensemble'] 
    path = [drs_root]
    for attr in attrs:
        if drs[attr] is None:
            val = '*'
        else:
            if attr == 'ensemble':
                val = 'r%di%dp%d' % drs.ensemble
            else:
                val = drs[attr]
        if val is None:
            break
        path.append(val)


    #!DEBUG
    assert len(path) == len(attrs)+1

    path = os.path.join(*path)
    log.debug('%s => %s' % (drs, repr(path)))
    return path




# BSD Licence
# Copyright (c) 2011, Science & Technology Facilities Council (STFC)
# All rights reserved.
#
# See the LICENSE file in the source distribution of this software for
# the full license text.


import tempfile
import shutil
import sys, os, re
from glob import glob
from StringIO import StringIO
import datetime

import gen_drs
from drslib.drs_tree import DRSTree
from drslib.drs import path_to_drs, drs_to_path, DRS
from drslib import config

from drs_tree_shared import TestEg, TestListing, test_dir



class TestListing1(TestListing):
    __test__ = True

    listing_file = 'realm_1.ls'

    def test_1(self):

        self._discover('MPI-M', 'ECHAM6-MPIOM-HR')
        pt = self.dt.pub_trees.values()[0]
        self._do_version(pt)

    def test_2(self):
        self._discover('MPI-M', 'ECHAM6-MPIOM-LR')
        pt = self.dt.pub_trees.values()[0]
        self._do_version(pt)

class TestListingOptVar(TestListing):
    __test__ = True

    listing_file = 'opt_var.ls'

    def test_1(self):
        # This should detect no pub_trees because realm was not specified
        self.dt.discover(self.incoming, activity='cmip5',
                         product='output1',
                         institute='MOHC',
                         model='HadGEM2-ES')
        assert len(self.dt.pub_trees) == 0


    def test_2(self):
        # This should accept the provided realm
        self.dt.discover(self.incoming, activity='cmip5',
                         product='output1',
                         institute='MOHC',
                         model='HadGEM2-ES',
                         realm='atmos')

        pt = self.dt.pub_trees.values()[0]
        self._do_version(pt)
        #!TODO: confirm realm set right


class TestCopyUpgrade(TestListing):
    """Test overriding the move command.
    """

    __test__ = True

    listing_file = 'realm_1.ls'

    def test_1(self):
        self._discover('MPI-M', 'ECHAM6-MPIOM-HR')
        self.dt.set_move_cmd('cp')
        
        pt = self.dt.pub_trees.values()[0]
        filename = self.dt.incoming[0][0]
        
        # Filename should remain after copy
        assert os.path.exists(filename)
        self._do_version(pt)
        assert os.path.exists(filename)
        
        # Check the filename did end up in the version
        for f, drs in pt.versions.values()[0]:
            if os.path.basename(f) == os.path.basename(filename):
                assert os.path.exists(f)
                break
        else:
            # No matching versioned file found
            assert False

class TestMapfile(TestListing):
    __test__ = True

    listing_file = 'realm_1.ls'

    def test_1(self):
        self._discover('MPI-M', 'ECHAM6-MPIOM-HR')
        pt = self.dt.pub_trees.values()[0]
        self._do_version(pt)

        # Make a mapfile
        fh = StringIO()
        pt.version_to_mapfile(self.today, fh)
        mapfile = fh.getvalue()


        print mapfile
        assert 'cmip5.output1.MPI-M.ECHAM6-MPIOM-HR.rcp45.mon.ocean.Omon.r1i1p1' in mapfile
        assert 'output1/MPI-M/ECHAM6-MPIOM-HR/rcp45/mon/ocean/Omon/r1i1p1/v%s'%self.today in mapfile


class TestThreeway(TestEg):
    __test__ = True

    listing_files = ['threeway_1.ls', 'threeway_2.ls', 'threeway_3.ls']

    def setUp(self):
        super(TestThreeway, self).setUp()

        self.dt = DRSTree(self.tmpdir)
        self.listing_iter = self._iterSetUpListings()

    def _iterSetUpListings(self):
        for listing_file in self.listing_files:
            listing_path = os.path.join(test_dir, listing_file)
            gen_drs.write_listing(self.tmpdir, listing_path)

            yield listing_path

    def _discover(self):
        self.dt.discover_incoming(self.incoming, activity='cmip5',
                         product='output1',
                         institute='MOHC',
                         model='HadGEM2-ES')

    def _do_version(self, pt, next_version):
        assert next_version not in pt.versions.keys()
        pt.do_version(next_version)
        assert next_version in pt.versions.keys()

    def _check_version(self, pt, version):
        for path, drs in pt.versions[version]:
            assert os.path.islink(path)
            # link is relative
            real_path = os.path.realpath(os.path.join(os.path.dirname(path),
                                                      os.readlink(path)))
            assert os.path.isfile(real_path)

            # Check variables match
            mo = re.search(r'/files/(.*?)_\d+/(.*?)_', real_path)
            assert mo.group(1) == mo.group(2)

    def test1(self):
        v = 1
        for listing_path in self.listing_iter:
            print 'Doing version %d' % v
            self._discover()
            assert len(self.dt.pub_trees) == 1
            pt = self.dt.pub_trees.values()[0]

            self._do_version(pt, v)
            self._check_version(pt, v)
            v += 1




#!/usr/bin/python
"""
initialise the drslib.p_cmip5 module.

"""

import os, sys
import xlrd, string, shelve
import re, glob
from whichdb import whichdb

import logging
log = logging.getLogger(__name__)

from drslib.config import table_path, table_path_csv

# Shelve version is designed to enable drslib to detect when the user needs to upgrade
# their shelves using "drstool init".  A value of 0 implies pre-versioning and is compatible
# with shelve versions up to this point.  A value greater than 0 requires the file SHELVE_VERSION_FILE
# to be present in the shelve directory containing the version.  If they don't match _find_shelves() will
# complain
SHELVE_VERSION = 1
SHELVE_VERSION_FILE = 'VERSION'

STANDARD_OUTPUT_XLS = 'standard_output_17Sep2010_mod.xls'
STANDARD_OUTPUT_XLS = 'standard_output_mod.xls'
TEMPLATE_XLS = 'CMIP5_archive_size_template.xls'
TEMPLATE_MAPPINGS = 'expt_id_mapping.txt'
TEMPLATE_SHELVE = 'template'
STDO_SHELVE = 'standard_output'
STDO_MIP_SHELVE = 'standard_output_mip'

re_cmor_mip = re.compile( 'variable_entry:(?P<var>.*?):::(?P<misc>.*?)dimensions:(?P<dims>.*?):::' )

CMOR_TABLE_DIR = table_path
CMOR_TABLE_CSV_DIR = table_path_csv
CMIP5_REQUEST_XLS ='/home/martin/python/cmip5/work2/esgf-drslib-p_cmip5-d324c7c/drslib/p_cmip5/xls/'

re_cmor_mip2 = re.compile( 'dimensions:(?P<dims>.*?):::' )

day_f10 = ['huss','tasmax','tasmin','tas','pr','psl','sfcWind','tossq','tos','omldamax']

def scan_table(mip,dir=CMOR_TABLE_DIR):
  ll = open( '%s/CMIP5_%s' % (dir, mip), 'r' ).readlines()

  lll = map( string.strip, ll )
  ssss = string.join( lll, ':::' )
  vitems = string.split( ssss, ':::variable_entry:' )[1:]

  ee = []
  for i in vitems:
    b1 = string.split( i, ':::')[0]
    var = string.strip( b1 )
    mm = re_cmor_mip2.findall( i )
    if len(mm) == 1:
      ds = string.split( string.strip(mm[0]) )
    elif len(mm) == 0:
      ds = 'scalar'
    else:
      log.warn(  'Mistake?? in scan_table %s' % str(mm) )
      ds = mm
      raise 'Mistake?? in scan_table %s' % str(mm)
    ee.append( (var,ds) )
  return tuple( ee )

def get_priority( v, lll ):
    kseg = 0
    for l in lll:
       if string.find( l, ',%s,' % v) != -1:
         return (string.split( l, ',' )[0], kseg)
       if l[0:8] == 'priority':
         kseg +=1
    return (None,None)

def scan_table_csv(mip,ee,dir=CMOR_TABLE_CSV_DIR):
  ll = open( os.path.join(dir, '%s.csv' % mip), 'r' ).readlines()

## strip out white space, so that variable can be identified unambiguously by ',var,'
  lll = map( lambda x: string.replace( string.strip(x), ' ',''), ll )
  vlist = map( lambda x: x[0], ee )

  ee2 = []
  for x in ee:
    p,kseg = get_priority( x[0], lll )
    if mip == 'day' and x[0] in day_f10:
      flag1 = 1
    else:
      flag1 = 0
    ee2.append( (x[0],x[1],p,flag1,kseg) )
  return tuple( ee2 )

usage = """usage: %prog [shelve-dir]

shelve-dir: Destination of data files of CMIP5 standard output and archive size.
"""
def _find_shelves(shelve_dir):
    """
    Return the location of CMIP5 shelve files as a dictionary.

    """

    _check_shelve_version(shelve_dir)

    # Locations of shelve files
    template = os.path.join(shelve_dir, TEMPLATE_SHELVE)
    stdo = os.path.join(shelve_dir, STDO_SHELVE)
    stdo_mip = os.path.join(shelve_dir, STDO_MIP_SHELVE)

    assert whichdb(template)
    assert whichdb(stdo)
    assert whichdb(stdo_mip)

    return dict(template=template, stdo=stdo, stdo_mip=stdo_mip)

def _check_shelve_version(shelve_dir):
    version_file = os.path.join(shelve_dir, SHELVE_VERSION_FILE)
    if os.path.exists(version_file):
        shelve_version = int(open(version_file).read().strip())
    else:
        shelve_version = 0

    log.info('Shelve version detected as %d' % shelve_version)

    if shelve_version != SHELVE_VERSION:
        raise Exception("Your shelve directory version is incompatible with this version of drslib"
                        "Please run 'drs_tool init' to reconstruct your shelves")

def init(shelve_dir,mip_dir,mip_csv_dir=None,xls_dir=None):
    """
    Create the shelve files needed to run p_cmip5.

    """

    if xls_dir == None:
      xls_dir = os.path.join(os.path.dirname(__file__), 'xls')
    if mip_csv_dir == None:
      if mip_dir[-1] == '/':
        mip_csv_dir = mip_dir[0:-1] + '_csv'
      else:
        mip_csv_dir = mip_dir + '_csv'

    stdo_xls = os.path.join(xls_dir, STANDARD_OUTPUT_XLS)
    template_xls = os.path.join(xls_dir, TEMPLATE_XLS)
    template_map = os.path.join(xls_dir, TEMPLATE_MAPPINGS)

    if not os.path.exists(shelve_dir):
        os.makedirs(shelve_dir)

    # Locations of shelve files
    template = os.path.join(shelve_dir, TEMPLATE_SHELVE)
    stdo = os.path.join(shelve_dir, STDO_SHELVE)
    stdo_mip = os.path.join(shelve_dir, STDO_MIP_SHELVE)

    ri = request_importer(template=template_xls, cmip5_stdo=stdo_xls)
    ri.import_template(template,expt_mapping_file=template_map)
    ri.import_standard_output(stdo)

    mi = mip_importer_rev(mip_dir,mip_csv_dir)
    mi.imprt(mip=stdo_mip)
    
    version_file = os.path.join(shelve_dir, SHELVE_VERSION_FILE)
    fh = open(version_file, 'w')
    print >>fh, SHELVE_VERSION
    fh.close()

#---------------------------------------------------------------------------
# Martin's code below this point with a few non-functional changes

def helper_year( val ):
  if type( val ) == type( 1. ):
    return int(val)
  elif type(val) in [type('x'), type(u'x')]:
    if string.strip(val) == '':
      return None
    if val[-1] == '*':
      val = val[:-1]
    return int(val)
  else:
    log.info(val)
    assert False, 'bad place to be'

class workflow:
  def __init__(self,name='fred',states=[]):
    self.statuses = {}
    for s in states:
      self.statuses[s] = ['all',None]
    self.name = name
    self.status = None
    self.ltypes = [type(()), type([])]


  def add(self, name, allowed='all', disallowed=None ):
    rv=1
    message = 'ok'
    if not( allowed in ['all',None] or type(allowed) in self.ltypes):
      rv = 0
      message = 'allowed: argument not valid'
    if not( disallowed in ['all',None] or type(disallowed) in self.ltypes):
      rv = 0
      message = 'disallowed: argument not valid'

    if rv == 0:
      return (rv,message)

    if (allowed,disallowed) in ( ['all',None],[None,'all'] ):
      rv=1
    elif (allowed == 'all' and disallowed != None) or (disallowed == 'all' and allowed != None):
      rv=0
      message = 'if allowed or disallowed is "all", the other must be None'
    elif type(allowed) in self.ltypes and type(disallowed) in self.ltypes:
      for a in allowed:
        if a in disallowed:
          rv = 0
          message = 'allowed and disallowed should not overlap'

    if rv != 0:
      self.statuses[name] = (allowed,disallowed)
    return (rv,message)

  def set(self,next):
    if self.allowed( next ):
      self.status = next
    else:
      log.info('%s %s' % (self.status, next))
      raise Exception('failed to set [%s] %s --> %s' % (self.name,self.status, next) )

  def reset(self):
    self.status = None

  def allowed(self,next):
    if next not in self.statuses.keys():
      return False

    if self.status == None:
      return True

    if self.statuses[self.status][1] == 'all':
      return False

    if self.statuses[self.status][0] == 'all':
      return True
 
    if type(self.statuses[self.status][0]) in self.ltypes:
      if next in self.statuses[self.status][0]:
         return True
  
    if type(self.statuses[self.status][1]) in self.ltypes:
      if next in self.statuses[self.status][1]:
         return False

    log.info('%s %s' % (self.status, next))
    log.info(self.statuses[self.status])
    raise Exception('cant determine whether allowed')
 

class request_importer:

  def __init__(self, template='CMIP5_template.xls', cmip5_stdo='xls' ):
    if not os.path.isfile(template):
      log.info('need a valid template file')
      raise Exception('bad arg')
    if not os.path.isfile(cmip5_stdo):
      log.info('need a valid standard output file')
      raise Exception('bad arg')

    self.template = template
    self.cmip5_stdo = cmip5_stdo

  def import_template(self,out='sh/template',expt_mapping_file='expt_id_mapping.txt'):
    mappings = {}
    assert os.path.isfile(expt_mapping_file), 'Cannot find expt_mapping_file %s' % expt_mapping_file
    ii = open(expt_mapping_file).readlines()
    for l in ii:
      bits = string.split( string.strip( l ) )
      mappings[bits[0]] = bits[1]
    
    sh = shelve.open( out, 'n' )
    book = xlrd.open_workbook( self.template )
    sheet = book.sheet_by_name('template')
    ydec = [1965, 1970, 1975, 1985, 1990, 1995, 2000, 2001, 2002, 2003, 2004, 2006, 2007, 2008, 2009, 2010]
    y30 = [1960, 1980, 2005]

    this_row = sheet.row(41)
    k0 = string.strip( str(this_row[10].value ) )
    assert k0 == 'decadalXXXX*', 'key should be decadalXXXX*: %s' % k0
    for y in ydec:
      key = 'decadal%4.4i' % y
      sh[key] = (41, '1.1',str(this_row[0].value), str(this_row[1].value))

    this_row = sheet.row(42)
    for y in y30:
      key = 'decadal%4.4i' % y
      sh[key] = (42, '1.2',str(this_row[0].value), str(this_row[1].value))

    this_row = sheet.row(49)
    k0 = string.strip( str(this_row[10].value ) )
    assert k0 == 'noVolcano', 'key should be noVolcano: %s' % k0
    for y in ydec + y30:
      key = 'noVolc%4.4i' % y
      sh[key] = (49, '1.3',str(this_row[0].value), str(this_row[1].value))

    histVars = ['Nat', 'Ant', 'GHG', 'SD', 'SI', 'SA', 'TO', 'SO', 'Oz', 'LU', 'Sl', 'Vl', 'SS', 'Ds', 'BC', 'MD', 'OC', 'AA']
    this_row = sheet.row(93)
    k0 = string.strip( str(this_row[10].value ) )
    assert k0 == 'historical???', 'key should be historical???: %s' % k0
    for v in histVars:
      if v not in ['GHG','Nat']:
        key = 'historical%s' % y
        sh[key] = (93, '7.3',str(this_row[0].value), str(this_row[1].value))

    rl1 = range(46,52) + range(56,97) + range(100,110)
    for r in rl1:
      this_row = sheet.row(r)
      key = string.strip( str(this_row[10].value ) )
      if key in mappings.keys():
        key = mappings[key]
      while key in sh.keys():
          key += '+'
      sh[key] = (r,str(this_row[2].value), str(this_row[0].value), str(this_row[1].value))
    sh.close()


  def import_standard_output(self, out='sh/standard_output'):
    sh = shelve.open( out, 'n' )
    book = xlrd.open_workbook( self.cmip5_stdo )
    self.import_other(sh,book)
    self.import_cfmip(sh,book)
    sh.close()

  def import_other(self,sh,book):
    oo = book.sheet_by_name('other output')
    rl1 = range(39,51) + range(54,95) + range(98,108)
    colh = ['M','N','Q','R']
    kkk = 0
    for col in [12,13,16,17]:
      ee = {}
      for r in rl1:
        iv = oo.row(r)[col].value
        if type(iv) == type(1.):
          ee[r] = ['list',('year',iv) ]
        ##if type(iv) in ( type('x'), "<type 'unicode'>"):
        else:
          item = string.strip(iv)
          if item == '':
             ee[r] = ('none',)
          elif item == 'all':
             ee[r] = ('all',)
          else:
             item = string.replace( item, '&', '' )
             item = string.replace( item, 'years', '' )
             item = string.replace( item, 'Years', '' )
             item = string.replace( item, 'Year', '' )
             item = string.replace( item, 'year', '' )
             item = string.replace( item, 'and', '' )
             item = string.replace( item, 'possibly', '' )
             item = string.strip( item )

             if string.find( item, 'eq') != -1:
               this_type = 'listrel'
               item = string.replace( item, 'eq', '' )
               log.info('listrel:: row %s [%s], col %s' % ( oo.row(r)[0].value, r, colh[kkk] ))
               log.info(item)
             else:
               this_type = 'list'
             bits = string.split( item, ',' )

             if string.find( item, 'corresp') != -1:
               assert r in [54,69], 'Only expecting corres in piControl r=%s' % r
               if r == 69 and colh[kkk] == 'M':
                 ll1 = []
                 for y in [1850, 1870, 1890, 1900,1910,1920,1930,1940, 1950, 1960, 1970, 1980,1990 , 2000, 2010, 2020, 2040, 2060, 2080, 2100]:
                   ll1.append( ('year',y) )
               else:
                 ll1 = { 'N':('slice',1986,2005), 'Q':('slice',1979,2008), 'R':('slice',111,140) }[colh[kkk]]
               ee[r] =  ['listrel',ll1]
               log.info('corres:: row %s, col %s' % ( oo.row(r)[0].value, colh[kkk] ))
             elif string.find( item, 'last') != -1:
               ee[r] =  ['corres',ll1]
             else:
               ll = [this_type]
               for b in bits:
                 if string.find( b, ':' ) != -1:
                   st,en,dd = map( int, string.split(b,':') )
                   for y in range(st,en+1,dd):
                     ll.append( ('year',y) )
                 elif string.find( b, '-' ) !=-1:
                   st,en = map( int, string.split(b,'-') )
                   ll.append( ('slice',st,en) )
                 else:
                   log.info('%s %s' % (bits,b) )
                   ll.append( ('year',int(b)) )
               ee[r] = ll
   
 
      sh[colh[kkk]] = ee
      kkk +=1
         
  def import_cfmip(self,sh,book):
    oo = book.sheet_by_name('CFMIP output')
    rl1 = range(6,27)
    ee = {}
#
# for each row, check for non-zero date ranges, and append to list. NB there may be multiple
# rows for each experiment.
#
    for r in rl1:
      this_row = oo.row(r)
      expt_nn = str(this_row[2].value)
      if string.strip( expt_nn ) != '':
        this_ee = ee.get( expt_nn, [ [],[],[],[] ] )
        for kseg in range(4):
          if str( this_row[3+kseg*2].value ) != '':
            y0 = helper_year( this_row[3+kseg*2].value )
            y9 = helper_year( this_row[4+kseg*2].value )
            if y0 != None:
                this_ee[kseg].append( (y0,y9) )
        ee[expt_nn] = this_ee
   
    log.info('putting cfmip in shelve')
    sh['cfmip'] = ee
         

##
## reads all user entered data into a dictionary.
##



class mip_importer_rev:

  def __init__(self,mip_dir,mip_csv_dir,mip_tables=None):
    self.input_mip_dir = mip_dir
    self.input_mip_csv_dir = mip_csv_dir
    assert os.path.isdir( self.input_mip_dir ), 'Specified input MIP table directory not found: %s' % self.input_mip_dir
    assert os.path.isdir( self.input_mip_csv_dir ), 'Specified input MIP csv directory not found: %s' % self.input_mip_csv_dir
    fl = glob.glob( self.input_mip_dir + '/CMIP5_*' )
    if mip_tables == None:
      self.mip_tables = []
      for f in fl:
        self.mip_tables.append( string.replace( string.split(f,'/')[-1], 'CMIP5_', '' ) )
    else:
      self.mip_tables = mip_tables
      for t in mip_tables:
         assert os.path.isfile( '%s/CMIP5_%s' % (self.input_mip_dir,t) ), 'File CMIP5_%s not found in %s' % (t,self.input_mip_dir)

  def imprt(self,mip='sh/standard_output_mip',return_as_dict=False):

    if return_as_dict:
      sh_mip = {}
    else:
      sh_mip = shelve.open( mip, 'n' )

    for t in self.mip_tables:
      ee = scan_table( t )
      if t == 'grids':
         ee2 = []
         for e in ee:
           ee2.append( (e[0],e[1], None, 0, None ) )
      else:
        ee2 = scan_table_csv(t,ee)
        if t == 'Omon':
          ee3 = scan_table_csv('Oyr',ee)
          ee4 = []
          for k in range(len(ee2)):
            assert len(ee2[k]) == 5, 'Bad ee2 element %s,%s' % (k, ee2[k] )
            if ee2[k][2] != None:
              ee4.append( ee2[k] )
            else:
              assert ee2[k][0] == ee3[k][0], 'mismatch at %s, %s, %s' % (k,str(ee2[k]),str(ee3[k]))
              ee4.append( (ee2[k][0],ee2[k][1], ee3[k][2], ee2[k][3], None ) )
          ee2 = ee4

      sh_mip[t] = ee2

    if return_as_dict:
      return sh_mip
    else:
      sh_mip.close()
      return None

if __name__ == '__main__':
  import sys
  if len( sys.argv ) != 3:
     print 'usage: init.py go <dir>'
     sys.exit(0)

  init( sys.argv[2],CMOR_TABLE_DIR, xls_dir=CMIP5_REQUEST_XLS )

  sh = shelve.open( sys.argv[2] + '/standard_output_mip', 'r' )
  tlist = ['cf3hr', 'cfSites', 'Oyr','Omon','aero','day','6hrPlev','3hr','cfMon']
  print sh.keys(), len(sh.keys())
  for t in tlist:
    l1 = sh[t]
    ee = scan_table( t )
    ee2 = scan_table_csv(t,ee)
    if t in ['Omon','cf3hr','cfSites']:
      if t == 'Omon':
         t2 = 'Oyr'
      else:
         t2 = 'Amon'
      ee3 = scan_table_csv(t2,ee)
      ee4 = []
      for k in range(len(ee2)):
        assert len(ee2[k]) == 5, 'Bad ee2 element %s,%s' % (k, ee2[k] )
        if ee2[k][2] != None:
          ee4.append( ee2[k] )
        else:
          assert ee2[k][0] == ee3[k][0], 'mismatch at %s, %s, %s' % (k,str(ee2[k]),str(ee3[k]))
          ee4.append( (ee2[k][0],ee2[k][1], ee3[k][2], ee2[k][3] ) )
      ee2 = ee4
      
    kkk = 0
    kk1 = 0
    for e in ee2:
      if e[2] == None:
        kkk += 1
      if e[3] == 1:
        kk1 += 1

    print t, len(ee), len(ee2), len(l1), kkk, kk1
    if kkk > 0:
       print '==================================='
       for e in ee2:
         if e[2] == None:
           print e
       print '==================================='
       
    if t == 'day' and kk1 != 10:
      print 'error',ee2[0:10]
    if t == 'day':
      print ee2[0:10]

    if len(ee) != len(l1):
      eevl = []
      for e in ee:
        eevl.append( e[0] )
      knf = 0
      kmd = []
      for i in range(len(l1)):
          rold = l1[i]
          vold = string.strip(rold[5])
          if vold not in eevl:
            print t,':: not found: ',vold
            knf += 1
          else:
            kmd.append( eevl.index( vold ) )
      knmd = []
      for i in range(len(eevl)):
           if i not in kmd:
             knmd.append( eevl[i] )
      knmd.sort()
      print knmd
    else:
      for i in range(len(l1)):
          rold = l1[i]
          rnew = ee[i]
          if rnew[0] != rold[5]:
            print t,i,rnew[0],rold[5]




# BSD Licence
# Copyright (c) 2010, Science & Technology Facilities Council (STFC)
# All rights reserved.
#
# See the LICENSE file in the source distribution of this software for
# the full license text.

"""
Tests compatible with nosetests

"""

import os

from drslib import cmip5
from drslib.drs import DRS

translator = cmip5.make_translator('cmip5')
translator_noversion = cmip5.make_translator('cmip5', with_version=False)

def get_drs1():
    return translator.filename_to_drs('tas_Amon_HadCM3_historicalNat_r1i1p1_185001-200512.nc')

def test_1():
    """
    Demonstrate succesfully parsing a filename

    """
    drs = get_drs1()

    print drs.__dict__

def test_2():
    """
    Demonstrate successfully generating a DRS path from a filename

    """
    drs = get_drs1()

    # Add the bits missing from the conversion
    drs.version = 2
    drs.product = 'output'
    
    path = translator.drs_to_filepath(drs)

    assert path=='cmip5/output/MOHC/HadCM3/historicalNat/mon/atmos/Amon/r1i1p1/v2/tas/tas_Amon_HadCM3_historicalNat_r1i1p1_185001-200512.nc'



def roundtrip_filename(filename):
    drs = translator.filename_to_drs(filename)

    drs.version = 1
    drs.product = 'output'
    assert drs.is_complete()

    path = translator.drs_to_filepath(drs)

    assert os.path.basename(path) == filename

def test_3():
    roundtrip_filename('tas_Amon_HadCM3_historicalNat_r1i1p1_185001-200512.nc')


def test_4():
    # Read some filenames from UKMO and roundtrip them
    fh = open(os.path.join(os.path.dirname(__file__), 'cmip5_test.ls'))
    for filename in fh:
        yield roundtrip_filename, filename.strip()

def test_5():
    # Regression test for a bug
    print translator.path_to_drs('cmip5/output/MOHC/HadCM3/historicalNat/mon/atmos/Amon/r1i1p1/v3/tas/tas_Amon_HadCM3_historicalNat_r1i1p1_185001-200512.nc')

def test_6():
    # Bug reported by Ag
    print translator.filename_to_drs('cct_Amon_HadGEM2-ES_piControl_r1i1p1.nc')


def test_7():
    # Test creating path without version

    drs = get_drs1()

    # Add the bits missing from the conversion
    drs.product = 'output'
    
    path = translator_noversion.drs_to_filepath(drs)

    assert path=='cmip5/output/MOHC/HadCM3/historicalNat/mon/atmos/tas/r1i1p1/tas_Amon_HadCM3_historicalNat_r1i1p1_185001-200512.nc'

def test_8():
    # Test files created from the cmor test suite (2010-06-25)
    # This list has been adapted from the result of running "make test"
    # in cmor.  Numeric prefixes to the "hfls" variable name have been removed.
    fh = open(os.path.join(os.path.dirname(__file__), 'cmor_test_ls'))
    for filename in fh:
        yield roundtrip_filename, filename.strip()

def test_9():
    # Regression test for bug in non-versioned path construction
    
    drs = get_drs1()

    # Add the bits missing from the conversion
    drs.product = 'output'
    
    path = translator_noversion.drs_to_path(drs)

    assert path=='cmip5/output/MOHC/HadCM3/historicalNat/mon/atmos/tas/r1i1p1'

def test_10():
    # Check instantiating DRS objects from other DRS objects

    drs = get_drs1()
    assert drs.version == None

    drs2 = DRS(drs, version=12)
    assert drs2.version == 12
    assert drs.model == drs2.model

def test_11():
    # Regression test for files in multiple realms

    drs = translator_noversion.filename_to_drs('snw_LImon_HadGEM2-ES_rcp45_r1i1p1_201512-204011.nc')

    assert drs.realm == 'landIce'

def test_11():
    # Check single ensemble number is allowed
    drs = translator_noversion.filename_to_drs('hur_Amon_HadGEM2-ES_historical_r1_186212-186311.nc')
    print drs.to_dataset_id()

def test_11():
    # Detect regression where 0-indexed ensembles are printed wrong
    drs = translator_noversion.filename_to_drs('areacella_fx_IPSL-CM5A-LR_piControl_r0i0p0.nc')
    assert drs.ensemble == (0, 0, 0)
    assert 'r0i0p0' in repr(drs)


def test_12():
    # Variables not in MIP table
    drs = translator_noversion.filename_to_drs('novar1_Amon_HadGEM2-ES_rcp45_r1i1p1_201512-204011.nc')


def test_13():
    # Variables with clim flag as in DRS doc
    drs = translator_noversion.filename_to_drs('novar1_Amon_HadGEM2-ES_rcp45_r1i1p1_201512-204011-clim.nc')
    assert drs.subset[2] == True

def test_14():
    # Variables with clim flag as produced by CMOR
    drs = translator_noversion.filename_to_drs('novar1_Amon_HadGEM2-ES_rcp45_r1i1p1_201512-204011_clim.nc')
    assert drs.subset[2] == True

def test_15():
    drs = translator.filepath_to_drs('cmip5/output1/IPSL/IPSL-CM5A-LR/piControl/3hr/atmos/3hr/r1i1p1/v20110324/ps/ps_3hr_IPSL-CM5A-LR_piControl_r1i1p1_275001010300-279912312100.nc')
    assert drs.institute == 'IPSL'

def test_16():
    # From Estani for Bug 111
    #ds='cmip5.output1.CSIRO-QCCCE.CSIRO-Mk3-6-0.1pctCO2.day.atmos.day.r1i1p1'
    fn='pr_day_CSIRO-Mk3-6-0_1pctCO2_r1i1p1_00010101-00201231.nc'
    path='cmip5/output1/CSIRO-QCCCE/CSIRO-Mk3-6-0/1pctCO2/day/atmos/day/r1i1p1/v20110518/pr'

    fullpath = os.path.join(path, fn)
    drs = translator.filepath_to_drs(fullpath)
    #drs.product, drs.institute, drs.model =  ds.split('.')[1:4]

    calculated_path = translator.drs_to_filepath(drs)
    assert calculated_path == fullpath

def test_17():
    # From Estani for bug 112
    drs = translator.filename_to_drs('psl_6hrPlev_MPI-ESM-LR_amip_r1i1p1_1979010100-197912311800.nc')
    print translator.drs_to_file(drs)

def test_18():
    # From Estani for bug 86
    fn = 'psl_6hrPlev_MPI-ESM-LR_amip_r1i1p1_1979010100-197912311800.nc'
    drs = translator.filename_to_drs(fn)
    calculated_fn = translator.drs_to_file(drs)

    assert calculated_fn == fn



# BSD Licence
# Copyright (c) 2011, Science & Technology Facilities Council (STFC)
# All rights reserved.
#
# See the LICENSE file in the source distribution of this software for
# the full license text.


from drslib.cmip5 import make_translator
from drslib import p_cmip5
from drslib import drs_tree
import gen_drs

import os, shutil
import metaconfig
import tempfile

LISTING = 'multi_product.ls'

config = metaconfig.get_config('drslib')
shelve_dir = config.get('p_cmip5', 'shelve-dir')
drs_root = '/cmip5'
vtrans = make_translator(drs_root)

def setup_module():
    global p_cmip5, listing, tmpdir, dt

    tmpdir = tempfile.mkdtemp(prefix='drs_tree-product-')

    print 'TMPDIR  ',tmpdir
    shelves = p_cmip5.init._find_shelves(shelve_dir)

    config_file = os.path.join(os.path.dirname(__file__), 'ukmo_sample.ini')
    listing = os.path.join(os.path.dirname(__file__), LISTING)
    gen_drs.write_listing(tmpdir, listing)
    
    p_cmip5 = p_cmip5.product.cmip5_product(mip_table_shelve=shelves['stdo_mip'],
                                            template=shelves['template'],
                                            stdo=shelves['stdo'],
                                            config=config_file,
                                            not_ok_excpt=True)

    dt = drs_tree.DRSTree(tmpdir)
    dt.set_p_cmip5(p_cmip5)


def test_product_dup():
    """
    Test scanning a set of files that are put into multiple products.

    """


    filenames = [f.strip() for f in open(listing).readlines()]
    def iter_files():
        for filename in filenames:
            yield filename, tmpdir


    dt.discover_incoming_fromfiles(iter_files())

    # There should be 2 pub-level datasets
    assert len(dt.pub_trees) == 2

    pt1, pt2 = dt.pub_trees.values()

    # They should be in separate products
    assert pt1.drs.product != pt2.drs.product

    #!NOTE: if the test fails before here it isn't really testing multi-product ingests
    #    probably the p_cmip5 algorithm has changed to put the test data into 1 product.

    # They should be disjoint
    set1 = set(x[0] for x in pt1._todo)
    set2 = set(x[0] for x in pt2._todo)
    assert set1.isdisjoint(set2)

    # Check the total number of files is right
    assert len(set1) + len(set2) == len(filenames)



def test_product_fixed():


    filenames = ['areacella_fx_IPSL-CM5A-LR_piControl_r0i0p0.nc']
    def iter_files():
        for filename in filenames:
            yield filename, tmpdir


    dt.discover_incoming_fromfiles(iter_files())


def teardown_module():
    shutil.rmtree(tmpdir)
    





"""
Test drslib is configured to accept decadal experiments with explicit years rather than "decadalXXXX"

"""

from drslib import cmip5
from drslib.translate import TranslationError

cmip5_trans = None

def setup():
    global cmip5_trans

    cmip5_trans = cmip5.make_translator('/cmip5', with_version=False)

years = range(1960, 2010)
        

def test():
    filepat = "zg_Amon_HadCM3_decadal%s_r4i2p1_196011-199012.nc"

    for year in years:
        filename = filepat % year

        print year, 

        try:
            drs_obj = cmip5_trans.filename_to_drs(filename)
        except TranslationError:
            raise AssertionError

        assert drs_obj.experiment == 'decadal%s' % year
    



# BSD Licence
# Copyright (c) 2011, Science & Technology Facilities Council (STFC)
# All rights reserved.
#
# See the LICENSE file in the source distribution of this software for
# the full license text.

"""
Generate mapfiles from streams of DRS objects

"""

#!TODO: check againsts similar code in datanode_admin and merge

import stat, os


def drs_to_id(drs):
    """
    Returns an esgpublish id for this drs object.

    The esgpublish id is a '.'-separated sequence of DRS components
    from the activity to realm level.
    
    """
    return '.'.join([drs.activity,
                     drs.product,
                     drs.institute,
                     drs.model,
                     drs.experiment,
                     drs.frequency,
                     drs.realm,
                     drs.table,
                     'r%di%dp%d' % drs.ensemble])

#!TODO: add callout to get parameters like checksum.
def write_mapfile(stream, fh):
    """
    Write an esgpublish mapfile from a stream of tuples (filepath, drs).

    """

    for path, drs in stream:
        file_stat = os.stat(path)
        size = file_stat[stat.ST_SIZE]
        mtime = file_stat[stat.ST_MTIME]

        print >>fh, ' | '.join([drs_to_id(drs), path, str(size), "mod_time=%f"%float(mtime)])
        
